# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import logging
from enum import IntEnum

from .util import CtxTypeError, ensure_dir, CLIError, color_map
from .events import EVENT_PARSER_GLOBAL_CREATE


CLI_LOGGER_NAME = 'cli'
# Add more logger names to this list so that ERROR, WARNING, INFO logs from these loggers can also be displayed
# without --debug flag.
cli_logger_names = [CLI_LOGGER_NAME]

LOG_FILE_ENCODING = 'utf-8'


class CliLogLevel(IntEnum):
    CRITICAL = 0
    ERROR = 1
    WARNING = 2
    INFO = 3
    DEBUG = 4


def get_logger(module_name=None):
    """ Get the logger for a module. If no module name is given, the current CLI logger is returned.

    Example:
        get_logger(__name__)

    :param module_name: The module to get the logger for
    :type module_name: str
    :return: The logger
    :rtype: logger
    """
    if module_name:
        logger_name = '{}.{}'.format(CLI_LOGGER_NAME, module_name)
    else:
        logger_name = CLI_LOGGER_NAME
    return logging.getLogger(logger_name)


class _CustomStreamHandler(logging.StreamHandler):

    @classmethod
    def wrap_with_color(cls, level_name, msg):
        color_marker = color_map[level_name.lower()]
        return '{}{}{}'.format(color_marker, msg, color_map['reset'])

    def __init__(self, log_level_config, log_format, enable_color):
        logging.StreamHandler.__init__(self)
        self.setLevel(log_level_config)
        self.enable_color = enable_color
        self.setFormatter(logging.Formatter(log_format))

    def format(self, record):
        msg = logging.StreamHandler.format(self, record)
        if self.enable_color:
            msg = self.wrap_with_color(record.levelname, msg)
        return msg


class CLILogging:  # pylint: disable=too-many-instance-attributes

    DEBUG_FLAG = '--debug'
    VERBOSE_FLAG = '--verbose'
    ONLY_SHOW_ERRORS_FLAG = '--only-show-errors'

    @staticmethod
    def on_global_arguments(_, **kwargs):
        arg_group = kwargs.get('arg_group')
        # The arguments for verbosity don't get parsed by argparse but we add it here for help.
        arg_group.add_argument(CLILogging.VERBOSE_FLAG, dest='_log_verbosity_verbose', action='store_true',
                               help='Increase logging verbosity. Use --debug for full debug logs.')
        arg_group.add_argument(CLILogging.DEBUG_FLAG, dest='_log_verbosity_debug', action='store_true',
                               help='Increase logging verbosity to show all debug logs.')
        arg_group.add_argument(CLILogging.ONLY_SHOW_ERRORS_FLAG, dest='_log_verbosity_only_show_errors',
                               action='store_true',
                               help='Only show errors, suppressing warnings.')

    def __init__(self, name, cli_ctx=None):
        """

        :param name: The name to be used for log files
        :type name: str
        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        """
        from .cli import CLI
        if cli_ctx is not None and not isinstance(cli_ctx, CLI):
            raise CtxTypeError(cli_ctx)
        self.logfile_name = '{}.log'.format(name)
        self.file_log_enabled = CLILogging._is_file_log_enabled(cli_ctx)
        self.log_dir = CLILogging._get_log_dir(cli_ctx)
        self.cli_ctx = cli_ctx
        self.cli_ctx.register_event(EVENT_PARSER_GLOBAL_CREATE, CLILogging.on_global_arguments)

        # The value is determined when `configure` is called
        self.log_level = None

    def configure(self, args):
        """ Configure the loggers with the appropriate log level etc.

        :param args: The arguments from the command line
        :type args: list
        """
        root_logger = logging.getLogger()

        if root_logger.handlers:
            # handlers already configured
            return

        self.log_level = self._determine_log_level(args)
        console_log_levels = self._get_console_log_levels()
        console_log_formats = self._get_console_log_formats()

        # Set the levels of the loggers to lowest level.
        # Handlers can override by choosing a higher level.
        root_logger.setLevel(logging.DEBUG)

        cli_loggers = [logging.getLogger(logger_name) for logger_name in cli_logger_names]
        for cli_logger in cli_loggers:
            cli_logger.setLevel(logging.DEBUG)
            cli_logger.propagate = False

        self._init_console_handlers(root_logger, cli_loggers, console_log_levels, console_log_formats)
        if self.file_log_enabled:
            self._init_logfile_handlers(root_logger, cli_loggers)
            get_logger(__name__).debug("File logging enabled - writing logs to '%s'.", self.log_dir)

    def _determine_log_level(self, args):
        """ Get verbose level by reading the arguments. """
        # arguments have higher precedence than config
        if CLILogging.ONLY_SHOW_ERRORS_FLAG in args:
            if CLILogging.DEBUG_FLAG in args or CLILogging.VERBOSE_FLAG in args:
                raise CLIError("--only-show-errors can't be used together with --debug or --verbose")
            self.cli_ctx.only_show_errors = True
            return CliLogLevel.ERROR
        if CLILogging.DEBUG_FLAG in args:
            self.cli_ctx.only_show_errors = False
            return CliLogLevel.DEBUG
        if CLILogging.VERBOSE_FLAG in args:
            self.cli_ctx.only_show_errors = False
            return CliLogLevel.INFO
        if self.cli_ctx.only_show_errors:
            # only_show_errors is enabled by config
            return CliLogLevel.ERROR
        return CliLogLevel.WARNING  # default to show WARNINGs and above

    def _init_console_handlers(self, root_logger, cli_loggers, log_levels, log_formats):
        root_logger.addHandler(_CustomStreamHandler(log_levels['root'],
                                                    log_formats['root'],
                                                    self.cli_ctx.enable_color))
        cli_logger_console_handler = _CustomStreamHandler(log_levels['cli'], log_formats['cli'],
                                                          self.cli_ctx.enable_color)
        for cli_logger in cli_loggers:
            cli_logger.addHandler(cli_logger_console_handler)

    def _init_logfile_handlers(self, root_logger, cli_loggers):
        ensure_dir(self.log_dir)
        log_file_path = os.path.join(self.log_dir, self.logfile_name)
        from logging.handlers import RotatingFileHandler
        logfile_handler = RotatingFileHandler(log_file_path, maxBytes=10 * 1024 * 1024, backupCount=5,
                                              encoding=LOG_FILE_ENCODING)
        lfmt = logging.Formatter('%(process)d : %(asctime)s : %(levelname)s : %(name)s : %(message)s')
        logfile_handler.setFormatter(lfmt)
        logfile_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(logfile_handler)
        for cli_logger in cli_loggers:
            cli_logger.addHandler(logfile_handler)

    @staticmethod
    def _is_file_log_enabled(cli_ctx):
        return cli_ctx.config.getboolean('logging', 'enable_log_file', fallback=False)

    @staticmethod
    def _get_log_dir(cli_ctx):
        default_dir = os.path.join(cli_ctx.config.config_dir, 'logs')
        return os.path.expanduser(cli_ctx.config.get('logging', 'log_dir', fallback=default_dir))

    def _get_console_log_levels(self):
        """Levels of cli logger and root logger for console logging.

        - cli logger level is controlled by the overall log level
        - root logger is only shown at DEBUG overall level
        """
        level_list = [
            # --only-show-critical [RESERVED]
            {
                'cli': logging.CRITICAL,
                'root': logging.CRITICAL
            },
            # --only-show-errors
            {
                'cli': logging.ERROR,
                'root': logging.CRITICAL
            },
            # (default)
            {
                'cli': logging.WARNING,
                'root': logging.CRITICAL,
            },
            # --verbose
            {
                'cli': logging.INFO,
                'root': logging.CRITICAL,
            },
            # --debug
            {
                'cli': logging.DEBUG,
                'root': logging.DEBUG,
            }]
        return level_list[self.log_level]

    def _get_console_log_formats(self):
        """Formats of cli logger and root logger for console logging, depending on color and level settings.

        - color:
          - True: Hide level names
          - False: Show level names (ERROR, WARNING, INFO, DEBUG)

        - level:
          - DEBUG: both cli and root logger names are shown
          - others: no logger names are shown (root logger won't be shown anyway)
        """

        elements = []

        if not self.cli_ctx.enable_color:
            elements.append('%(levelname)s')

        if self.log_level == CliLogLevel.DEBUG:
            elements.append('%(name)s')

        elements.append('%(message)s')

        log_format = ': '.join(elements)
        return {
            # Even though these loggers use the same format, keep the dict for further tuning
            'cli': log_format,
            'root': log_format
        }



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from unittest import mock

from knack.events import EVENT_PARSER_GLOBAL_CREATE
from knack.query import CLIQuery
from tests.util import MockContext


class TestQueryEventHandling(unittest.TestCase):

    def setUp(self):
        self.mock_ctx = MockContext()
        self.cli_query = CLIQuery(cli_ctx=self.mock_ctx)

    def test_cli_ctx_type_error(self):
        with self.assertRaises(TypeError):
            CLIQuery(cli_ctx=object())

    def test_query_argument_registrations(self):
        parser_arg_group_mock = mock.MagicMock()
        self.mock_ctx.raise_event(EVENT_PARSER_GLOBAL_CREATE, arg_group=parser_arg_group_mock)
        parser_arg_group_mock.add_argument.assert_any_call('--query',
                                                           metavar=mock.ANY,
                                                           dest=mock.ANY,
                                                           help=mock.ANY,
                                                           type=mock.ANY)


class TestQuery(unittest.TestCase):
    '''Tests for the values that can be passed to the --query parameter.
    These tests ensure that we are handling invalid queries correctly and raising appropriate errors
    that argparse can then handle.
    (We are not testing JMESPath itself here)
    '''

    def test_query_valid_1(self):
        query = 'length(@)'
        # Should not raise any exception as it is valid
        CLIQuery.jmespath_type(query)

    def test_query_valid_2(self):
        query = "[?propertyX.propertyY.propertyZ=='AValue'].[col1,col2]"
        # Should not raise any exception as it is valid
        CLIQuery.jmespath_type(query)

    def test_query_empty(self):
        query = ''
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)

    def test_query_unbalanced(self):
        query = 'length(@'
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)

    def test_query_invalid_1(self):
        query = '[?asdf=asdf]'
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)

    def test_query_invalid_2(self):
        query = '[?name=My Value]'
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)

    def test_query_invalid_3(self):
        query = "[].location='westus'"
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)

    def test_query_invalid_4(self):
        query = "length([?contains('id', 'Publishers'])"
        with self.assertRaises(ValueError):
            CLIQuery.jmespath_type(query)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from unittest import mock

import sys
import argparse

from knack.arguments import ArgumentsContext
from knack.commands import CLICommandsLoader, CommandGroup

from tests.util import DummyCLI, redirect_io, assert_in_multi_line


def example_handler(arg1, arg2=None, arg3=None):
    """ Short summary here. Long summary here. Still long summary. """
    pass


def example_arg_handler(arg1, opt1, arg2=None, opt2=None, arg3=None,
                        opt3=None, arg4=None, opt4=None, arg5=None, opt5=None):
    pass


class TestCommandExperimental(unittest.TestCase):

    def setUp(self):

        from knack.help_files import helps

        class ExperimentalTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super(ExperimentalTestCommandLoader, self).load_command_table(args)
                with CommandGroup(self, '', '{}#{{}}'.format(__name__)) as g:
                    g.command('cmd1', 'example_handler', is_experimental=True)

                with CommandGroup(self, 'grp1', '{}#{{}}'.format(__name__), is_experimental=True) as g:
                    g.command('cmd1', 'example_handler')

                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, '') as c:
                    c.argument('arg1', options_list=['--arg', '-a'], required=False, type=int, choices=[1, 2, 3])
                    c.argument('arg2', options_list=['-b'], required=True, choices=['a', 'b', 'c'])

                super(ExperimentalTestCommandLoader, self).load_arguments(command)

        helps['grp1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=ExperimentalTestCommandLoader)

    @redirect_io
    def test_experimental_command_implicitly_execute(self):
        """ Ensure general warning displayed when running command from an experimental parent group. """
        self.cli_ctx.invoke('grp1 cmd1 -b b'.split())
        actual = self.io.getvalue()
        expected = "Command group 'grp1' is experimental and under development."
        self.assertIn(expected, actual)

    @redirect_io
    def test_experimental_command_group_help(self):
        """ Ensure experimental commands appear correctly in group help view. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('-h'.split())
        actual = self.io.getvalue()
        expected = """
Group
    {}

Subgroups:
    grp1 [Experimental] : A group.

Commands:
    cmd1 [Experimental] : Short summary here.

""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    def test_experimental_command_plain_execute(self):
        """ Ensure general warning displayed when running experimental command. """
        self.cli_ctx.invoke('cmd1 -b b'.split())
        actual = self.io.getvalue()
        expected = "This command is experimental and under development."
        self.assertIn(expected, actual)


class TestCommandGroupExperimental(unittest.TestCase):

    def setUp(self):

        from knack.help_files import helps

        class ExperimentalTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super(ExperimentalTestCommandLoader, self).load_command_table(args)

                with CommandGroup(self, 'group1', '{}#{{}}'.format(__name__), is_experimental=True) as g:
                    g.command('cmd1', 'example_handler')

                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, '') as c:
                    c.argument('arg1', options_list=['--arg', '-a'], required=False, type=int, choices=[1, 2, 3])
                    c.argument('arg2', options_list=['-b'], required=True, choices=['a', 'b', 'c'])

                super(ExperimentalTestCommandLoader, self).load_arguments(command)

        helps['group1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=ExperimentalTestCommandLoader)

    @redirect_io
    def test_experimental_command_group_help_plain(self):
        """ Ensure help warnings appear for experimental command group help. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('group1 -h'.split())
        actual = self.io.getvalue()
        expected = """
Group
    cli group1 : A group.
        This command group is experimental and under development.
Commands:
    cmd1 : Short summary here.

""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    def test_experimental_command_implicitly(self):
        """ Ensure help warning displayed for command in experimental because of a experimental parent group. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('group1 cmd1 -h'.split())
        actual = self.io.getvalue()
        expected = """
Command
    {} group1 cmd1 : Short summary here.
        Long summary here. Still long summary.
        Command group 'group1' is experimental and under development.
""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)


class TestArgumentExperimental(unittest.TestCase):

    def setUp(self):
        from knack.help_files import helps

        class LoggerAction(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                print("Side-effect from some original action!", file=sys.stderr)

        class ExperimentalTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super(ExperimentalTestCommandLoader, self).load_command_table(args)
                with CommandGroup(self, '', '{}#{{}}'.format(__name__)) as g:
                    g.command('arg-test', 'example_arg_handler')
                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, 'arg-test') as c:
                    c.argument('arg1', help='Arg1', is_experimental=True, action=LoggerAction)

                super(ExperimentalTestCommandLoader, self).load_arguments(command)

        helps['grp1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=ExperimentalTestCommandLoader)

    @redirect_io
    def test_experimental_arguments_command_help(self):
        """ Ensure experimental arguments appear correctly in command help view. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('arg-test -h'.split())
        actual = self.io.getvalue()
        expected = """
Arguments
    --arg1 [Experimental] [Required] : Arg1.
        Argument '--arg1' is experimental and under development.
""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    def test_experimental_arguments_execute(self):
        """ Ensure deprecated arguments can be used. """
        self.cli_ctx.invoke('arg-test --arg1 foo --opt1 bar'.split())
        actual = self.io.getvalue()
        experimental_expected = "Argument '--arg1' is experimental and under development."
        self.assertIn(experimental_expected, actual)

        action_expected = "Side-effect from some original action!"
        self.assertIn(action_expected, actual)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import os
import stat
import configparser

from .util import ensure_dir

_UNSET = object()

CONFIG_FILE_ENCODING = 'utf-8'


def get_config_parser():
    return configparser.ConfigParser()  # keep this for backward compatibility


class CLIConfig(object):
    _BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                       '0': False, 'no': False, 'false': False, 'off': False}

    _DEFAULT_CONFIG_ENV_VAR_PREFIX = 'CLI'
    _DEFAULT_CONFIG_DIR = os.path.expanduser(os.path.join('~', '.{}'.format('cli')))
    _DEFAULT_CONFIG_FILE_NAME = 'config'
    _CONFIG_DEFAULTS_SECTION = 'defaults'

    def __init__(self, config_dir=None, config_env_var_prefix=None, config_file_name=None, use_local_config=None):
        """ Manages configuration options available in the CLI

        :param config_dir: The directory to store config files
        :type config_dir: str
        :param config_env_var_prefix: The prefix for config environment variables
        :type config_env_var_prefix: str
        :param config_file_name: The name given to the config file to be created
        :type config_file_name: str
        """
        config_dir = config_dir or CLIConfig._DEFAULT_CONFIG_DIR
        ensure_dir(config_dir)
        config_env_var_prefix = config_env_var_prefix or CLIConfig._DEFAULT_CONFIG_ENV_VAR_PREFIX
        env_var_prefix = '{}_'.format(config_env_var_prefix.upper())
        default_config_dir = os.path.expanduser(config_dir)
        self.config_dir = os.environ.get('{}CONFIG_DIR'.format(env_var_prefix), default_config_dir)
        configuration_file_name = config_file_name or CLIConfig._DEFAULT_CONFIG_FILE_NAME
        self.config_path = os.path.join(self.config_dir, configuration_file_name)
        self._env_var_format = '{}{}'.format(env_var_prefix, '{section}_{option}')
        self.defaults_section_name = CLIConfig._CONFIG_DEFAULTS_SECTION
        self.use_local_config = use_local_config
        self._config_file_chain = []

        current_dir = None
        try:
            current_dir = os.getcwd()
        except FileNotFoundError:
            from .log import get_logger
            logger = get_logger()
            logger.warning("The working directory has been deleted or recreated. "
                           "Local config is ignored.")

        config_dir_name = os.path.basename(self.config_dir)
        while current_dir:
            current_config_dir = os.path.join(current_dir, config_dir_name)
            # Stop if already in the default .azure
            if (os.path.normcase(os.path.normpath(current_config_dir)) ==
                    os.path.normcase(os.path.normpath(self.config_dir))):
                break
            if os.path.isdir(current_config_dir):
                self._config_file_chain.append(_ConfigFile(current_config_dir,
                                                           os.path.join(current_config_dir, configuration_file_name)))
            # Stop if already in root drive
            if current_dir == os.path.dirname(current_dir):
                break
            current_dir = os.path.dirname(current_dir)
        self._config_file_chain.append(_ConfigFile(self.config_dir, self.config_path))

    def env_var_name(self, section, option):
        return self._env_var_format.format(section=section.upper(),
                                           option=option.upper())

    def has_option(self, section, option):
        if self.env_var_name(section, option) in os.environ:
            return True
        config_files = self._config_file_chain if self.use_local_config else self._config_file_chain[-1:]
        return bool(next((f for f in config_files if f.has_option(section, option)), False))

    def get(self, section, option, fallback=_UNSET):
        env = self.env_var_name(section, option)
        if env in os.environ:
            return os.environ[env]
        last_ex = None
        for config in self._config_file_chain if self.use_local_config else self._config_file_chain[-1:]:
            try:
                return config.get(section, option)
            except (configparser.NoSectionError, configparser.NoOptionError) as ex:
                last_ex = ex

        if fallback is _UNSET:
            raise last_ex  # pylint:disable=raising-bad-type
        return fallback

    def sections(self):
        combined_sections = []
        # Go through the config chain and combine all sections
        for config in self._config_file_chain if self.use_local_config else self._config_file_chain[-1:]:
            sections = config.sections()
            for section in sections:
                if section not in combined_sections:
                    combined_sections.append(section)
        return combined_sections

    def items(self, section):
        import re
        # Only allow valid env vars, in all caps: CLI_SECTION_TEST_OPTION, CLI_SECTION__TEST_OPTION
        pattern = self.env_var_name(section, '([0-9A-Z_]+)')
        env_entries = []
        for k in os.environ:
            # Must be a full match, otherwise CLI_SECTION_T part in CLI_MYSECTION_Test_Option will match
            matched = re.fullmatch(pattern, k)
            if matched:
                # (name, value, ENV_VAR_NAME)
                item = (matched.group(1).lower(), os.environ[k], k)
                env_entries.append(item)

        # Prepare result with env entries first
        result = {c[0]: c for c in env_entries}
        # Add entries from config files if they do not exist yet
        for config in self._config_file_chain if self.use_local_config else self._config_file_chain[-1:]:
            try:
                entries = config.items(section)
                for name, value in entries:
                    if name not in result:
                        result[name] = (name, value, config.config_path)
            except (configparser.NoSectionError, configparser.NoOptionError):
                pass
        return [{'name': name, 'value': value, 'source': source} for name, value, source in result.values()]

    def getint(self, section, option, fallback=_UNSET):
        return int(self.get(section, option, fallback))

    def getfloat(self, section, option, fallback=_UNSET):
        return float(self.get(section, option, fallback))

    def getboolean(self, section, option, fallback=_UNSET):
        val = str(self.get(section, option, fallback))
        if val.lower() not in CLIConfig._BOOLEAN_STATES:
            raise ValueError('Not a boolean: {}'.format(val))
        return CLIConfig._BOOLEAN_STATES[val.lower()]

    def set_value(self, section, option, value):
        if self.use_local_config:
            current_config_dir = os.path.join(os.getcwd(), os.path.basename(self.config_dir))
            config_file_path = os.path.join(current_config_dir, os.path.basename(self.config_path))
            if config_file_path == self._config_file_chain[0].config_path:
                self._config_file_chain[0].set_value(section, option, value)
            else:
                config = _ConfigFile(current_config_dir, config_file_path)
                config.set_value(section, option, value)
                self._config_file_chain.insert(0, config)
        else:
            self._config_file_chain[-1].set_value(section, option, value)

    def set_to_use_local_config(self, use_local_config):
        self.use_local_config = use_local_config

    def remove_option(self, section, option):
        for config in self._config_file_chain if self.use_local_config else self._config_file_chain[-1:]:
            if config.remove_option(section, option):
                return True
        return False


class _ConfigFile(object):
    _BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                       '0': False, 'no': False, 'false': False, 'off': False}

    def __init__(self, config_dir, config_path, config_comment=None):
        """ Manage configuration options available in the CLI

        :param config_dir: The directory to store the config file
        :type config_dir: str
        :param config_path: The path of the config file
        :type config_path: str
        :param config_comment: The comment which will be written into the head of the config file
        :type config_comment: str

        When 'config_comment' is given, each line should start with # or ;. For details about INI file comment,
        see https://docs.python.org/3/library/configparser.html#supported-ini-file-structure
        """
        self.config_dir = config_dir
        self.config_path = config_path
        self.config_comment = config_comment
        self.config_parser = configparser.ConfigParser()
        if os.path.exists(config_path):
            self.config_parser.read(config_path, encoding=CONFIG_FILE_ENCODING)

    def items(self, section):
        return self.config_parser.items(section) if self.config_parser else []

    def sections(self):
        return self.config_parser.sections() if self.config_parser else []

    def has_option(self, section, option):
        return self.config_parser.has_option(section, option) if self.config_parser else False

    def get(self, section, option):
        if self.config_parser:
            return self.config_parser.get(section, option)
        raise configparser.NoOptionError(option, section)

    def getint(self, section, option):
        return int(self.get(section, option))

    def getfloat(self, section, option):
        return float(self.get(section, option))

    def getboolean(self, section, option):
        val = str(self.get(section, option))
        if val.lower() not in _ConfigFile._BOOLEAN_STATES:
            raise ValueError('Not a boolean: {}'.format(val))
        return _ConfigFile._BOOLEAN_STATES[val.lower()]

    def set(self, config):
        ensure_dir(self.config_dir)
        with open(self.config_path, 'w', encoding=CONFIG_FILE_ENCODING) as configfile:
            if self.config_comment:
                configfile.write(self.config_comment + '\n')
            config.write(configfile)
        os.chmod(self.config_path, stat.S_IRUSR | stat.S_IWUSR)
        self.config_parser.read(self.config_path)

    def set_value(self, section, option, value):
        config = configparser.ConfigParser()
        config.read(self.config_path)
        try:
            config.add_section(section)
        except configparser.DuplicateSectionError:
            pass
        config.set(section, option, value)
        self.set(config)

    def remove_option(self, section, option):
        existed = False
        if self.config_parser:
            try:
                existed = self.config_parser.remove_option(section, option)
                self.set(self.config_parser)
            except configparser.NoSectionError:
                pass
        return existed

    def remove_section(self, section):
        if self.config_parser and self.config_parser.remove_section(section):
            self.set(self.config_parser)
            return True
        return False

    def clear(self):
        if self.config_parser:
            for section in self.config_parser.sections():
                self.config_parser.remove_section(section)
            self.set(self.config_parser)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from unittest import mock

import sys
import argparse

from knack.arguments import ArgumentsContext
from knack.commands import CLICommandsLoader, CommandGroup

from tests.util import DummyCLI, redirect_io, assert_in_multi_line, disable_color


def example_handler(arg1, arg2=None, arg3=None):
    """ Short summary here. Long summary here. Still long summary. """
    pass


def example_arg_handler(arg1, opt1, arg2=None, opt2=None, arg3=None,
                        opt3=None, arg4=None, opt4=None, arg5=None, opt5=None):
    pass


class TestCommandPreview(unittest.TestCase):

    def setUp(self):

        from knack.help_files import helps

        class PreviewTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super().load_command_table(args)
                with CommandGroup(self, '', '{}#{{}}'.format(__name__)) as g:
                    g.command('cmd1', 'example_handler', is_preview=True)

                with CommandGroup(self, 'grp1', '{}#{{}}'.format(__name__), is_preview=True) as g:
                    g.command('cmd1', 'example_handler')

                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, '') as c:
                    c.argument('arg1', options_list=['--arg', '-a'], required=False, type=int, choices=[1, 2, 3])
                    c.argument('arg2', options_list=['-b'], required=True, choices=['a', 'b', 'c'])

                super().load_arguments(command)

        helps['grp1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=PreviewTestCommandLoader)

    @redirect_io
    def test_preview_command_group_help(self):
        """ Ensure preview commands appear correctly in group help view. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('-h'.split())
        actual = self.io.getvalue()
        expected = """
Group
    {}

Subgroups:
    grp1 [Preview] : A group.

Commands:
    cmd1 [Preview] : Short summary here.

""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    def test_preview_command_plain_execute(self):
        """ Ensure general warning displayed when running preview command. """
        self.cli_ctx.invoke('cmd1 -b b'.split())
        actual = self.io.getvalue()
        expected = "This command is in preview. It may be changed/removed in a future release."
        self.assertIn(expected, actual)

    @redirect_io
    def test_preview_command_plain_execute_only_show_error(self):
        """ Ensure warning is suppressed when running preview command. """
        # Directly use --only-show-errors
        self.cli_ctx.invoke('cmd1 -b b --only-show-errors'.split())
        actual = self.io.getvalue()
        self.assertNotIn("preview", actual)

        # Apply --only-show-errors with config
        self.cli_ctx.only_show_errors = True
        self.cli_ctx.config.set_value('core', 'only_show_errors', 'True')
        self.cli_ctx.invoke('cmd1 -b b'.split())
        actual = self.io.getvalue()
        self.assertNotIn("preview", actual)
        self.cli_ctx.config.set_value('core', 'only_show_errors', '')
        self.cli_ctx.only_show_errors = False

    @redirect_io
    @disable_color
    def test_preview_command_plain_execute_no_color(self):
        """ Ensure warning is displayed without color. """
        self.cli_ctx.invoke('cmd1 -b b'.split())
        actual = self.io.getvalue()
        self.assertIn("WARNING: This command is in preview. It may be changed/removed in a future release.", actual)

    @redirect_io
    def test_preview_command_implicitly_execute(self):
        """ Ensure general warning displayed when running command from a preview parent group. """
        self.cli_ctx.invoke('grp1 cmd1 -b b'.split())
        actual = self.io.getvalue()
        expected = "Command group 'grp1' is in preview. It may be changed/removed in a future release."
        self.assertIn(expected, actual)

    @redirect_io
    @disable_color
    def test_preview_command_implicitly_no_color(self):
        """ Ensure warning is displayed without color. """
        self.cli_ctx.invoke('grp1 cmd1 -b b'.split())
        actual = self.io.getvalue()
        expected = "WARNING: Command group 'grp1' is in preview. It may be changed/removed in a future release."
        self.assertIn(expected, actual)


class TestCommandGroupPreview(unittest.TestCase):

    def setUp(self):

        from knack.help_files import helps

        class PreviewTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super().load_command_table(args)

                with CommandGroup(self, 'group1', '{}#{{}}'.format(__name__), is_preview=True) as g:
                    g.command('cmd1', 'example_handler')

                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, '') as c:
                    c.argument('arg1', options_list=['--arg', '-a'], required=False, type=int, choices=[1, 2, 3])
                    c.argument('arg2', options_list=['-b'], required=True, choices=['a', 'b', 'c'])

                super().load_arguments(command)

        helps['group1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=PreviewTestCommandLoader)

    @redirect_io
    def test_preview_command_group_help_plain(self):
        """ Ensure help warnings appear for preview command group help. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('group1 -h'.split())
        actual = self.io.getvalue()
        expected = """
Group
    cli group1 : A group.
        This command group is in preview. It may be changed/removed in a future release.
Commands:
    cmd1 : Short summary here.

""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    @disable_color
    def test_preview_command_group_help_plain_no_color(self):
        """ Ensure warning is displayed without color. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('group1 -h'.split())
        actual = self.io.getvalue()
        expected = """
Group
    cli group1 : A group.
        WARNING: This command group is in preview. It may be changed/removed in a future release.
Commands:
    cmd1 : Short summary here.

""".format(self.cli_ctx.name)
        self.assertEqual(expected, actual)

    @redirect_io
    def test_preview_command_implicitly(self):
        """ Ensure help warning displayed for command in preview because of a preview parent group. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('group1 cmd1 -h'.split())
        actual = self.io.getvalue()
        expected = """
Command
    {} group1 cmd1 : Short summary here.
        Long summary here. Still long summary.
        Command group 'group1' is in preview. It may be changed/removed in a future
        release.
""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)


class TestArgumentPreview(unittest.TestCase):

    def setUp(self):
        from knack.help_files import helps

        class LoggerAction(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                print("Side-effect from some original action!", file=sys.stderr)

        class PreviewTestCommandLoader(CLICommandsLoader):
            def load_command_table(self, args):
                super().load_command_table(args)
                with CommandGroup(self, '', '{}#{{}}'.format(__name__)) as g:
                    g.command('arg-test', 'example_arg_handler')
                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, 'arg-test') as c:
                    c.argument('arg1', help='Arg1', is_preview=True, action=LoggerAction)

                super().load_arguments(command)

        helps['grp1'] = """
    type: group
    short-summary: A group.
"""
        self.cli_ctx = DummyCLI(commands_loader_cls=PreviewTestCommandLoader)

    @redirect_io
    def test_preview_arguments_command_help(self):
        """ Ensure preview arguments appear correctly in command help view. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('arg-test -h'.split())
        actual = self.io.getvalue()
        expected = """
Arguments
    --arg1 [Preview] [Required] : Arg1.
        Argument '--arg1' is in preview. It may be changed/removed in a future release.
""".format(self.cli_ctx.name)
        assert_in_multi_line(expected, actual)

    @redirect_io
    @disable_color
    def test_preview_arguments_command_help_no_color(self):
        """ Ensure warning is displayed without color. """
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('arg-test -h'.split())
        actual = self.io.getvalue()
        expected = """
Arguments
    --arg1 [Preview] [Required] : Arg1.
        WARNING: Argument '--arg1' is in preview. It may be changed/removed in a future release.
""".format(self.cli_ctx.name)
        self.assertIn(expected, actual)

    @redirect_io
    def test_preview_arguments_execute(self):
        """ Ensure deprecated arguments can be used. """
        self.cli_ctx.invoke('arg-test --arg1 foo --opt1 bar'.split())
        actual = self.io.getvalue()
        preview_expected = "Argument '--arg1' is in preview. It may be changed/removed in a future release."
        self.assertIn(preview_expected, actual)

        action_expected = "Side-effect from some original action!"
        self.assertIn(action_expected, actual)

    @redirect_io
    @disable_color
    def test_preview_arguments_execute_no_color(self):
        """ Ensure warning is displayed without color. """
        self.cli_ctx.invoke('arg-test --arg1 foo --opt1 bar'.split())
        actual = self.io.getvalue()
        preview_expected = "WARNING: Argument '--arg1' is in preview. It may be changed/removed in a future release."
        self.assertIn(preview_expected, actual)

        action_expected = "Side-effect from some original action!"
        self.assertIn(action_expected, actual)


    @redirect_io
    def test_preview_arguments_execute_only_show_error(self):
        """ Ensure warning is suppressed when using preview arguments. """
        self.cli_ctx.invoke('arg-test --arg1 foo --opt1 bar --only-show-errors'.split())
        actual = self.io.getvalue()
        self.assertNotIn("preview", actual)

        action_expected = "Side-effect from some original action!"
        self.assertIn(action_expected, actual)

if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import types
from collections import OrderedDict, defaultdict
from importlib import import_module

from .deprecation import Deprecated
from .preview import PreviewItem
from .experimental import ExperimentalItem
from .prompting import prompt_y_n, NoTTYException
from .util import CLIError, CtxTypeError
from .arguments import ArgumentRegistry, CLICommandArgument
from .introspection import extract_args_from_signature, extract_full_summary_from_signature
from .events import (EVENT_CMDLOADER_LOAD_COMMAND_TABLE, EVENT_CMDLOADER_LOAD_ARGUMENTS,
                     EVENT_COMMAND_CANCELLED)
from .log import get_logger
from .validators import DefaultInt, DefaultStr

logger = get_logger(__name__)


PREVIEW_EXPERIMENTAL_CONFLICT_ERROR = "Failed to register {} '{}', " \
                                      "is_preview and is_experimental can't be true at the same time"


class CLICommand(object):  # pylint:disable=too-many-instance-attributes

    # pylint: disable=unused-argument
    def __init__(self, cli_ctx, name, handler, description=None, table_transformer=None,
                 arguments_loader=None, description_loader=None,
                 formatter_class=None, deprecate_info=None, validator=None, confirmation=None, preview_info=None,
                 experimental_info=None, **kwargs):
        """ The command object that goes into the command table.

        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        :param name: The name of the command (e.g. 'mygroup mycommand')
        :type name: str
        :param handler: The function that will handle this command
        :type handler: function
        :param description: The description for the command
        :type description: str
        :param table_transformer: A function that transforms the command output for displaying in a table
        :type table_transformer: function
        :param arguments_loader: The function that defines how the arguments for the command should be loaded
        :type arguments_loader: function
        :param description_loader: The function that defines how the description for the command should be loaded
        :type description_loader: function
        :param formatter_class: The formatter for how help should be displayed
        :type formatter_class: class
        :param deprecate_info: Deprecation message to display when this command is invoked
        :type deprecate_info: str
        :param preview_info: Indicates a command is in preview
        :type preview_info: bool
        :param experimental_info: Indicates a command is experimental
        :type experimental_info: bool
        :param validator: The command validator
        :param confirmation: User confirmation required for command
        :type confirmation: bool, str, callable
        :param kwargs: Extra kwargs that are currently ignored
        """
        from .cli import CLI
        if cli_ctx is not None and not isinstance(cli_ctx, CLI):
            raise CtxTypeError(cli_ctx)
        self.cli_ctx = cli_ctx
        self.name = name
        self.handler = handler
        self.help = None
        self.description = description_loader if description_loader and self.should_load_description() else description
        self.arguments = {}
        self.arguments_loader = arguments_loader
        self.table_transformer = table_transformer
        self.formatter_class = formatter_class
        self.deprecate_info = deprecate_info
        self.preview_info = preview_info
        self.experimental_info = experimental_info
        self.confirmation = confirmation
        self.validator = validator

    def should_load_description(self):
        return not self.cli_ctx.data['completer_active']

    def _resolve_default_value_from_config_file(self, arg, overrides):
        default_key = overrides.settings.get('configured_default', None)
        if not default_key:
            return

        defaults_section = self.cli_ctx.config.defaults_section_name
        use_local_config_original = self.cli_ctx.config.use_local_config
        self.cli_ctx.config.set_to_use_local_config(True)
        config_value = self.cli_ctx.config.get(defaults_section, default_key, None)
        self.cli_ctx.config.set_to_use_local_config(use_local_config_original)
        if config_value:
            logger.info("Configured default '%s' for arg %s", config_value, arg.name)
            overrides.settings['default'] = DefaultStr(config_value)
            overrides.settings['required'] = False
            overrides.settings['default_value_source'] = 'Config'

    def load_arguments(self):
        if self.arguments_loader:
            cmd_args = self.arguments_loader()
            if self.confirmation:
                cmd_args.append(('yes',
                                 CLICommandArgument(dest='yes', options_list=['--yes', '-y'],
                                                    action='store_true', help='Do not prompt for confirmation.')))
            self.arguments.update(cmd_args)

    def add_argument(self, param_name, *option_strings, **kwargs):
        dest = kwargs.pop('dest', None)
        argument = CLICommandArgument(dest or param_name, options_list=option_strings, **kwargs)
        self.arguments[param_name] = argument

    def update_argument(self, param_name, argtype):
        arg = self.arguments[param_name]
        # resolve defaults from either environment variable or config file
        self._resolve_default_value_from_config_file(arg, argtype)
        arg.type.update(other=argtype)
        arg_default = arg.type.settings.get('default', None)
        # apply DefaultStr and DefaultInt to allow distinguishing between
        # when a default was applied or when the user specified a value
        # that coincides with the default
        if isinstance(arg_default, str):
            arg_default = DefaultStr(arg_default)
        elif isinstance(arg_default, int) and not isinstance(arg_default, bool):
            # bool's is_default is ignored for now, because
            # 1. bool is a subclass of int, so isinstance(True, int) cannot distinguish between int and bool
            # 2. bool is not extendable according to https://stackoverflow.com/a/2172204/2199657
            arg_default = DefaultInt(arg_default)
        # update the default
        if arg_default:
            arg.type.settings['default'] = arg_default

    def execute(self, **kwargs):
        return self(**kwargs)

    def __call__(self, *args, **kwargs):

        cmd_args = args[0]

        confirm = self.confirmation and not cmd_args.pop('yes', None) \
            and not self.cli_ctx.config.getboolean('core', 'disable_confirm_prompt', fallback=False)

        if confirm and not self._user_confirmed(self.confirmation, cmd_args):
            self.cli_ctx.raise_event(EVENT_COMMAND_CANCELLED, command=self.name, command_args=cmd_args)
            raise CLIError('Operation cancelled.')
        return self.handler(*args, **kwargs)

    @staticmethod
    def _user_confirmed(confirmation, command_args):
        if callable(confirmation):
            return confirmation(command_args)
        try:
            if isinstance(confirmation, str):
                return prompt_y_n(confirmation)
            return prompt_y_n('Are you sure you want to perform this operation?')
        except NoTTYException:
            logger.warning('Unable to prompt for confirmation as no tty available. Use --yes.')
            return False


# pylint: disable=too-many-instance-attributes
class CLICommandsLoader(object):

    def __init__(self, cli_ctx=None, command_cls=CLICommand, excluded_command_handler_args=None):
        """ The loader of commands. It contains the command table and argument registries.

        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        :param command_cls: The command type that the command table will be populated with
        :type command_cls: knack.commands.CLICommand
        :param excluded_command_handler_args: List of params to ignore and not extract from a commands handler.
                                              By default we ignore ['self', 'kwargs'].
        :type excluded_command_handler_args: list of str
        """
        from .cli import CLI
        if cli_ctx is not None and not isinstance(cli_ctx, CLI):
            raise CtxTypeError(cli_ctx)
        self.cli_ctx = cli_ctx
        self.command_cls = command_cls
        self.skip_applicability = False
        self.excluded_command_handler_args = excluded_command_handler_args
        # A command table is a dictionary of name -> CLICommand instances
        self.command_table = {}
        # A command group table is a dictionary of names -> CommandGroup instances
        self.command_group_table = {}
        # An argument registry stores all arguments for commands
        self.argument_registry = ArgumentRegistry()
        self.extra_argument_registry = defaultdict(lambda: {})

    def _populate_command_group_table_with_subgroups(self, name):
        if not name:
            return

        # ensure all subgroups have some entry in the command group table
        name_components = name.split()
        for i, _ in enumerate(name_components):
            subgroup_name = ' '.join(name_components[:i + 1])
            if subgroup_name not in self.command_group_table:
                self.command_group_table[subgroup_name] = {}

    def load_command_table(self, args):  # pylint: disable=unused-argument
        """ Load commands into the command table

        :param args: List of the arguments from the command line
        :type args: list
        :return: The ordered command table
        :rtype: collections.OrderedDict
        """
        self.cli_ctx.raise_event(EVENT_CMDLOADER_LOAD_COMMAND_TABLE, cmd_tbl=self.command_table)
        return OrderedDict(self.command_table)

    def load_arguments(self, command):
        """ Load the arguments for the specified command

        :param command: The command to load arguments for
        :type command: str
        """
        from knack.arguments import ArgumentsContext

        self.cli_ctx.raise_event(EVENT_CMDLOADER_LOAD_ARGUMENTS, cmd_tbl=self.command_table, command=command)
        try:
            self.command_table[command].load_arguments()
        except KeyError:
            return

        # ensure global 'cmd' is ignored
        with ArgumentsContext(self, '') as c:
            c.ignore('cmd')

        self._apply_parameter_info(command, self.command_table[command])

    def _apply_parameter_info(self, command_name, command):
        for argument_name in command.arguments:
            overrides = self.argument_registry.get_cli_argument(command_name, argument_name)
            command.update_argument(argument_name, overrides)
        # Add any arguments explicitly registered for this command
        for argument_name, argument_definition in self.extra_argument_registry[command_name].items():
            command.arguments[argument_name] = argument_definition
            command.update_argument(argument_name, self.argument_registry.get_cli_argument(command_name, argument_name))

    def create_command(self, name, operation, **kwargs):
        """ Constructs the command object that can then be added to the command table """
        if not isinstance(operation, str):
            raise ValueError("Operation must be a string. Got '{}'".format(operation))

        name = ' '.join(name.split())

        client_factory = kwargs.get('client_factory', None)

        def _command_handler(command_args):
            op = CLICommandsLoader._get_op_handler(operation)
            client = client_factory(command_args) if client_factory else None
            result = op(client, **command_args) if client else op(**command_args)
            return result

        def arguments_loader():
            return list(extract_args_from_signature(CLICommandsLoader._get_op_handler(operation),
                                                    excluded_params=self.excluded_command_handler_args))

        def description_loader():
            return extract_full_summary_from_signature(CLICommandsLoader._get_op_handler(operation))

        kwargs['arguments_loader'] = arguments_loader
        kwargs['description_loader'] = description_loader

        cmd = self.command_cls(self.cli_ctx, name, _command_handler, **kwargs)
        return cmd

    @staticmethod
    def _get_op_handler(operation):
        """ Import and load the operation handler """
        try:
            mod_to_import, attr_path = operation.split('#')
            op = import_module(mod_to_import)
            for part in attr_path.split('.'):
                op = getattr(op, part)
            if isinstance(op, types.FunctionType):
                return op
            # op as types.MethodType
            return op.__func__
        except (ValueError, AttributeError) as ex:
            raise ValueError("The operation '{}' is invalid.".format(operation)) from ex

    def deprecate(self, **kwargs):
        kwargs['object_type'] = 'command group'
        return Deprecated(self.cli_ctx, **kwargs)


class CommandGroup(object):

    def __init__(self, command_loader, group_name, operations_tmpl, **kwargs):
        """ Context manager for registering commands that share common properties.

        :param command_loader: The command loader that commands will be registered into
        :type command_loader: knack.commands.CLICommandsLoader
        :param group_name: The name of the group of commands in the command hierarchy
        :type group_name: str
        :param operations_tmpl: The template for handlers for this group of commands (e.g. '__main__#{}')
        :type operations_tmpl: str
        :param kwargs: Kwargs to apply to all commands in this group.
                       Possible values: `client_factory`, `arguments_loader`, `description_loader`, `description`,
                       `formatter_class`, `table_transformer`, `deprecate_info`, `validator`, `confirmation`.
        """
        self.command_loader = command_loader
        self.group_name = group_name
        self.operations_tmpl = operations_tmpl
        self.group_kwargs = kwargs
        Deprecated.ensure_new_style_deprecation(self.command_loader.cli_ctx, self.group_kwargs, 'command group')
        if kwargs['deprecate_info']:
            kwargs['deprecate_info'].target = group_name

        is_preview = kwargs.get('is_preview', False)
        is_experimental = kwargs.get('is_experimental', False)
        if is_preview and is_experimental:
            raise CLIError(PREVIEW_EXPERIMENTAL_CONFLICT_ERROR.format("command group", group_name))
        if is_preview:
            kwargs['preview_info'] = PreviewItem(
                cli_ctx=self.command_loader.cli_ctx,
                target=group_name,
                object_type='command group'
            )
        if is_experimental:
            kwargs['experimental_info'] = ExperimentalItem(
                cli_ctx=self.command_loader.cli_ctx,
                target=group_name,
                object_type='command group'
            )
        command_loader._populate_command_group_table_with_subgroups(group_name)  # pylint: disable=protected-access
        self.command_loader.command_group_table[group_name] = self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def command(self, name, handler_name, **kwargs):
        """ Register a command into the command table

        :param name: The name of the command
        :type name: str
        :param handler_name: The name of the handler that will be applied to the operations template
        :type handler_name: str
        :param kwargs: Kwargs to apply to the command.
                       Possible values: `client_factory`, `arguments_loader`, `description_loader`, `description`,
                       `formatter_class`, `table_transformer`, `deprecate_info`, `validator`, `confirmation`,
                       `is_preview`, `is_experimental`.
        """
        import copy

        command_name = '{} {}'.format(self.group_name, name) if self.group_name else name
        command_kwargs = copy.deepcopy(self.group_kwargs)
        command_kwargs.update(kwargs)

        # don't inherit deprecation, preview and experimental info from command group
        # https://github.com/Azure/azure-cli/blob/683b9709b67c4c9e8df92f9fbd53cbf83b6973d3/src/azure-cli-core/azure/cli/core/commands/__init__.py#L1155
        command_kwargs['deprecate_info'] = kwargs.get('deprecate_info', None)

        is_preview = kwargs.get('is_preview', False)
        is_experimental = kwargs.get('is_experimental', False)
        if is_preview and is_experimental:
            raise CLIError(PREVIEW_EXPERIMENTAL_CONFLICT_ERROR.format("command", self.group_name + " " + name))

        command_kwargs['preview_info'] = None
        if is_preview:
            command_kwargs['preview_info'] = PreviewItem(self.command_loader.cli_ctx, object_type='command')
        command_kwargs['experimental_info'] = None
        if is_experimental:
            command_kwargs['experimental_info'] = ExperimentalItem(self.command_loader.cli_ctx, object_type='command')

        self.command_loader._populate_command_group_table_with_subgroups(' '.join(command_name.split()[:-1]))  # pylint: disable=protected-access
        self.command_loader.command_table[command_name] = self.command_loader.create_command(
            command_name,
            self.operations_tmpl.format(handler_name),
            **command_kwargs)

    def deprecate(self, **kwargs):
        kwargs['object_type'] = 'command'
        return Deprecated(self.command_loader.cli_ctx, **kwargs)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

EVENT_CLI_PRE_EXECUTE = 'Cli.PreExecute'
EVENT_CLI_SUCCESSFUL_EXECUTE = 'Cli.SuccessfulExecute'
EVENT_CLI_POST_EXECUTE = 'Cli.PostExecute'

EVENT_INVOKER_PRE_CMD_TBL_CREATE = 'CommandInvoker.OnPreCommandTableCreate'
EVENT_INVOKER_POST_CMD_TBL_CREATE = 'CommandInvoker.OnPostCommandTableCreate'
EVENT_INVOKER_CMD_TBL_LOADED = 'CommandInvoker.OnCommandTableLoaded'
EVENT_INVOKER_PRE_PARSE_ARGS = 'CommandInvoker.OnPreParseArgs'
EVENT_INVOKER_POST_PARSE_ARGS = 'CommandInvoker.OnPostParseArgs'
EVENT_INVOKER_TRANSFORM_RESULT = 'CommandInvoker.OnTransformResult'
EVENT_INVOKER_FILTER_RESULT = 'CommandInvoker.OnFilterResult'

EVENT_PARSER_GLOBAL_CREATE = 'CommandParser.OnGlobalArgumentsCreate'

EVENT_CMDLOADER_LOAD_COMMAND_TABLE = 'CommandLoader.OnLoadCommandTable'
EVENT_CMDLOADER_LOAD_ARGUMENTS = 'CommandLoader.OnLoadArguments'

EVENT_COMMAND_CANCELLED = 'Command.OnOperationCancelled'



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from io import StringIO

from knack.parser import CLICommandParser
from knack.commands import CLICommand
from knack.arguments import enum_choice_list
from tests.util import MockContext, redirect_io


class TestParser(unittest.TestCase):

    def setUp(self):
        self.io = StringIO()
        self.mock_ctx = MockContext()

    def tearDown(self):
        self.io.close()

    def test_register_simple_commands(self):
        def test_handler1():
            pass

        def test_handler2():
            pass

        command = CLICommand(self.mock_ctx, 'command the-name', test_handler1)
        command2 = CLICommand(self.mock_ctx, 'sub-command the-second-name', test_handler2)
        cmd_table = {'command the-name': command, 'sub-command the-second-name': command2}
        self.mock_ctx.commands_loader.command_table = cmd_table

        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)
        args = parser.parse_args('command the-name'.split())
        self.assertIs(args.func, command)

        args = parser.parse_args('sub-command the-second-name'.split())
        self.assertIs(args.func, command2)

        CLICommandParser.error = VerifyError(self,)
        parser.parse_args('sub-command'.split())
        self.assertTrue(CLICommandParser.error.called)

    def test_required_parameter(self):
        def test_handler(args):  # pylint: disable=unused-argument
            pass

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('req', '--req', required=True)
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table

        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)

        args = parser.parse_args('test command --req yep'.split())
        self.assertIs(args.func, command)

        CLICommandParser.error = VerifyError(self)
        parser.parse_args('test command'.split())
        self.assertTrue(CLICommandParser.error.called)

    def test_nargs_parameter(self):
        def test_handler():
            pass

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('req', '--req', required=True, nargs=2)
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table

        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)

        args = parser.parse_args('test command --req yep nope'.split())
        self.assertIs(args.func, command)

        CLICommandParser.error = VerifyError(self)
        parser.parse_args('test command -req yep'.split())
        self.assertTrue(CLICommandParser.error.called)

    def _enum_parser(self):
        from enum import Enum

        class TestEnum(Enum):  # pylint: disable=too-few-public-methods

            opt1 = "ALL_CAPS"
            opt2 = "camelCase"
            opt3 = "snake_case"

        def test_handler():
            pass

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('opt', '--opt', required=True, **enum_choice_list(TestEnum))
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table

        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)
        return parser

    def test_case_insensitive_enum_choices(self):
        parser = self._enum_parser()
        args = parser.parse_args('test command --opt alL_cAps'.split())
        self.assertEqual(args.opt, 'ALL_CAPS')

        args = parser.parse_args('test command --opt CAMELCASE'.split())
        self.assertEqual(args.opt, 'camelCase')

        args = parser.parse_args('test command --opt sNake_CASE'.split())
        self.assertEqual(args.opt, 'snake_case')

    @redirect_io
    def test_check_value_invalid_command(self):
        parser = self._enum_parser()
        with self.assertRaises(SystemExit) as cm:
            parser.parse_args('test command1'.split())  # 'command1' is invalid
        actual = self.io.getvalue()
        assert "is not in the" in actual and "command group" in actual

    @redirect_io
    def test_check_value_invalid_argument_value(self):
        parser = self._enum_parser()
        with self.assertRaises(SystemExit) as cm:
            parser.parse_args('test command --opt foo'.split())  # 'foo' is invalid
        actual = self.io.getvalue()
        assert "is not a valid value for" in actual

    def test_cli_ctx_type_error(self):
        with self.assertRaises(TypeError):
            CLICommandParser(cli_ctx=object())

    def test_extra_nonargparse_parameters(self):
        """ Add argument that has non argparse parameters.

            'mycustomarg' should be filtered out and load_command_table
            should complete successfully instead of throwing
            TypeError: __init__() got an unexpected keyword argument 'mycustomarg'
        """
        def test_handler():
            pass

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('req', '--req', required=True, mycustomarg=True)
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table
        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)

    def test_prefix_file_expansion(self):
        import json, os

        def test_handler():
            pass

        def create_test_file(file, contents):
            with open(file, 'w') as f:
                f.write(contents)

        def remove_test_file(file):
            os.remove(file)

        json_test_data = json.dumps({'one': 1, 'two': 2, 'three': 3})
        create_test_file('test.json', json_test_data)

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('json_data', '--param')
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table
        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)

        args = parser.parse_args('test command --param @test.json'.split())
        self.assertEqual(json_test_data, args.json_data)

        remove_test_file('test.json')

    def test_help_string_with_literal_percent_does_not_crash(self):
        def test_handler():
            pass

        command = CLICommand(self.mock_ctx, 'test command', test_handler)
        command.add_argument('date_fmt', '--date-fmt', help='Expected format: %Y-%m-%d')
        cmd_table = {'test command': command}
        self.mock_ctx.commands_loader.command_table = cmd_table

        parser = CLICommandParser()
        parser.load_command_table(self.mock_ctx.commands_loader)

    def test_help_string_preserves_argparse_placeholders(self):
        sanitized = CLICommandParser._sanitize_help_for_argparse('default is %(default)s (100%% expected)')
        self.assertEqual(sanitized, 'default is %(default)s (100%% expected)')


class VerifyError(object):  # pylint: disable=too-few-public-methods

    def __init__(self, test, substr=None):
        self.test = test
        self.substr = substr
        self.called = False

    def __call__(self, message):
        if self.substr:
            self.test.assertGreaterEqual(message.find(self.substr), 0)
        self.called = True


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from collections import namedtuple
from datetime import date, time, datetime
from unittest import mock

from knack.util import todict, to_snake_case, is_modern_terminal


class TestUtils(unittest.TestCase):

    def test_application_todict_none(self):
        the_input = None
        actual = todict(the_input)
        expected = None
        self.assertEqual(actual, expected)

    def test_application_todict_dict_empty(self):
        the_input = {}
        actual = todict(the_input)
        expected = {}
        self.assertEqual(actual, expected)

    def test_application_todict_dict(self):
        the_input = {'a': 'b'}
        actual = todict(the_input)
        expected = {'a': 'b'}
        self.assertEqual(actual, expected)

    def test_application_todict_list(self):
        the_input = [{'a': 'b'}]
        actual = todict(the_input)
        expected = [{'a': 'b'}]
        self.assertEqual(actual, expected)

    def test_application_todict_obj(self):
        MyObject = namedtuple('MyObject', 'a b')
        the_input = MyObject('x', 'y')
        actual = todict(the_input)
        expected = {'a': 'x', 'b': 'y'}
        self.assertEqual(actual, expected)

    def test_application_todict_dict_with_obj(self):
        MyObject = namedtuple('MyObject', 'a b')
        mo = MyObject('x', 'y')
        the_input = {'a': mo}
        actual = todict(the_input)
        expected = {'a': {'a': 'x', 'b': 'y'}}
        self.assertEqual(actual, expected)

    def test_application_todict_dict_with_date(self):
        the_input = date(2017, 10, 13)
        actual = todict(the_input)
        expected = the_input.isoformat()
        self.assertEqual(actual, expected)

    def test_application_todict_dict_with_datetime(self):
        the_input = datetime(2017, 10, 13, 1, 23, 45)
        actual = todict(the_input)
        expected = the_input.isoformat()
        self.assertEqual(actual, expected)

    def test_application_todict_dict_with_time(self):
        the_input = time(1, 23, 45)
        actual = todict(the_input)
        expected = the_input.isoformat()
        self.assertEqual(actual, expected)

    def test_to_snake_case_from_camel(self):
        the_input = 'thisIsCamelCase'
        expected = 'this_is_camel_case'
        actual = to_snake_case(the_input)
        self.assertEqual(expected, actual)

    def test_to_snake_case_empty(self):
        the_input = ''
        expected = ''
        actual = to_snake_case(the_input)
        self.assertEqual(expected, actual)

    def test_to_snake_case_already_snake(self):
        the_input = 'this_is_snake_cased'
        expected = 'this_is_snake_cased'
        actual = to_snake_case(the_input)
        self.assertEqual(expected, actual)

    def test_is_modern_terminal(self):
        with mock.patch.dict("os.environ", clear=True):
            self.assertEqual(is_modern_terminal(), False)
        with mock.patch.dict("os.environ", TERM_PROGRAM='vscode'):
            self.assertEqual(is_modern_terminal(), True)
        with mock.patch.dict("os.environ", PYCHARM_HOSTED='1'):
            self.assertEqual(is_modern_terminal(), True)
        with mock.patch.dict("os.environ", WT_SESSION='c25cb945-246a-49e5-b37a-1e4b6671b916'):
            self.assertEqual(is_modern_terminal(), True)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import argparse
from collections import defaultdict

from .deprecation import Deprecated
from .preview import PreviewItem
from .experimental import ExperimentalItem
from .log import get_logger
from .util import CLIError, status_tag_messages


logger = get_logger(__name__)


class CLIArgumentType(object):

    REMOVE = '---REMOVE---'

    def __init__(self, overrides=None, **kwargs):
        """A base CLI Argument Type that can be applied to multiple command arguments

        :param overrides: The base argument that you are overriding
        :type overrides: knack.arguments.CLIArgumentType
        :param kwargs: Possible values: `options_list`, `validator`, `completer`, `nargs`, `action`, `const`, `default`,
                       `type`, `choices`, `required`, `help`, `metavar`. See /docs/arguments.md.
        """
        if isinstance(overrides, str):
            raise ValueError("Overrides has to be a {} (cannot be a string)".format(CLIArgumentType.__name__))
        options_list = kwargs.get('options_list', None)
        if options_list and isinstance(options_list, str):
            kwargs['options_list'] = [options_list]
        self.settings = {}
        self.update(overrides, **kwargs)

    def update(self, other=None, **kwargs):
        if other:
            self.settings.update(**other.settings)
        self.settings.update(**kwargs)


class CLICommandArgument(object):

    NAMED_ARGUMENTS = ['options_list', 'validator', 'completer', 'arg_group', 'deprecate_info', 'preview_info',
                       'experimental_info', 'default_value_source']

    def __init__(self, dest=None, argtype=None, **kwargs):
        """An argument that has a specific destination parameter.

        :param dest: The parameter that this argument is for
        :type dest: str
        :param argtype: The argument type for this command argument
        :type argtype: knack.arguments.CLIArgumentType
        :param kwargs: see knack.arguments.CLIArgumentType
        """
        self.type = CLIArgumentType(overrides=argtype, **kwargs)
        if dest:
            self.type.update(dest=dest)

        # We'll do an early fault detection to find any instances where we have inconsistent
        # set of parameters for argparse
        if not self.options.get('dest', False):
            raise ValueError('Missing dest')
        if not self.options_list:  # pylint: disable=access-member-before-definition
            self.options_list = ('--{}'.format(self.options['dest'].replace('_', '-')),)

    def __getattr__(self, name):
        if name in self.NAMED_ARGUMENTS:
            return self.type.settings.get(name, None)
        if name == 'name':
            return self.type.settings.get('dest', None)
        if name == 'options':
            return {key: value for key, value in self.type.settings.items()
                    if key != 'options' and key not in self.NAMED_ARGUMENTS and
                    not value == CLIArgumentType.REMOVE}
        if name == 'choices':
            return self.type.settings.get(name, None)

        raise AttributeError(name)

    def __setattr__(self, name, value):  # pylint: disable=inconsistent-return-statements
        if name == 'type':
            return super().__setattr__(name, value)
        self.type.settings[name] = value


class ArgumentRegistry(object):
    """A registry of all the arguments registered"""

    def __init__(self):
        self.arguments = defaultdict(lambda: {})

    def register_cli_argument(self, scope, dest, argtype, **kwargs):
        """ Add an argument to the argument registry

        :param scope: The command level to apply the argument registration (e.g. 'mygroup mycommand')
        :type scope: str
        :param dest: The parameter/destination that this argument is for
        :type dest: str
        :param argtype: The argument type for this command argument
        :type argtype: knack.arguments.CLIArgumentType
        :param kwargs: see knack.arguments.CLIArgumentType
        """
        argument = CLIArgumentType(overrides=argtype, **kwargs)
        self.arguments[scope][dest] = argument

    def get_cli_argument(self, command, name):
        """ Get the argument for the command after applying the scope hierarchy

        :param command: The command that we want the argument for
        :type command: str
        :param name: The name of the argument
        :type name: str
        :return: The CLI command after all overrides in the scope hierarchy have been applied
        :rtype: knack.arguments.CLIArgumentType
        """
        parts = command.split()
        result = CLIArgumentType()
        for index in range(0, len(parts) + 1):
            probe = ' '.join(parts[0:index])
            override = self.arguments.get(probe, {}).get(name, None)
            if override:
                result.update(override)
        return result


class ArgumentsContext(object):
    def __init__(self, command_loader, command_scope, **kwargs):  # pylint: disable=unused-argument
        """ Context manager to register arguments

        :param command_loader: The command loader that arguments should be registered into
        :type command_loader: knack.commands.CLICommandsLoader
        :param command_scope: The scope to which arguments in this context apply.
                              More specific scopes will override less specific scopes in the event of a conflict.
        :type command_scope: str
        """
        self.command_loader = command_loader
        self.command_scope = command_scope
        self.is_stale = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.is_stale = True

    def _applicable(self):
        if self.command_loader.skip_applicability:
            return True
        return self.command_loader.cli_ctx.invocation.data['command_string'].startswith(self.command_scope)

    def _check_stale(self):
        if self.is_stale:
            message = "command authoring error: argument context '{}' is stale! " \
                      "Check that the subsequent block for has a corresponding `as` " \
                      "statement.".format(self.command_scope)
            logger.error(message)
            raise CLIError(message)

    def _get_parent_class(self, **kwargs):
        # wrap any existing action
        action = kwargs.get('action', None)
        parent_class = argparse.Action

        # action is either a user-defined Action class or a string referring a library-defined Action
        if isinstance(action, type) and issubclass(action, argparse.Action):
            parent_class = action
        elif isinstance(action, str):
            parent_class = self.command_loader.cli_ctx.invocation.parser._registries['action'][action]  # pylint: disable=protected-access
        return parent_class

    def _handle_deprecations(self, argument_dest, **kwargs):

        def _handle_argument_deprecation(deprecate_info):

            parent_class = self._get_parent_class(**kwargs)

            class DeprecatedArgumentAction(parent_class):

                def __call__(self, parser, namespace, values, option_string=None):
                    if not hasattr(namespace, '_argument_deprecations'):
                        setattr(namespace, '_argument_deprecations', [deprecate_info])
                    else:
                        namespace._argument_deprecations.append(deprecate_info)  # pylint: disable=protected-access
                    try:
                        super().__call__(parser, namespace, values, option_string)
                    except NotImplementedError:
                        setattr(namespace, self.dest, values)

            return DeprecatedArgumentAction

        def _handle_option_deprecation(deprecated_options):

            if not isinstance(deprecated_options, list):
                deprecated_options = [deprecated_options]

            parent_class = self._get_parent_class(**kwargs)

            class DeprecatedOptionAction(parent_class):

                def __call__(self, parser, namespace, values, option_string=None):
                    deprecated_opt = next((x for x in deprecated_options if option_string == x.target), None)
                    if deprecated_opt:
                        if not hasattr(namespace, '_argument_deprecations'):
                            setattr(namespace, '_argument_deprecations', [deprecated_opt])
                        else:
                            namespace._argument_deprecations.append(deprecated_opt)  # pylint: disable=protected-access
                    try:
                        super().__call__(parser, namespace, values, option_string)
                    except NotImplementedError:
                        setattr(namespace, self.dest, values)

            return DeprecatedOptionAction

        action = kwargs.get('action', None)

        deprecate_info = kwargs.get('deprecate_info', None)
        if deprecate_info:
            deprecate_info.target = deprecate_info.target or argument_dest
            action = _handle_argument_deprecation(deprecate_info)
        deprecated_opts = [x for x in kwargs.get('options_list', []) if isinstance(x, Deprecated)]
        if deprecated_opts:
            action = _handle_option_deprecation(deprecated_opts)
        return action

    def _handle_previews(self, argument_dest, **kwargs):

        if not kwargs.get('is_preview', False):
            return kwargs

        def _handle_argument_preview(preview_info):

            parent_class = self._get_parent_class(**kwargs)

            class PreviewArgumentAction(parent_class):

                def __call__(self, parser, namespace, values, option_string=None):
                    if not hasattr(namespace, '_argument_previews'):
                        setattr(namespace, '_argument_previews', [preview_info])
                    else:
                        namespace._argument_previews.append(preview_info)  # pylint: disable=protected-access
                    try:
                        super().__call__(parser, namespace, values, option_string)
                    except NotImplementedError:
                        setattr(namespace, self.dest, values)

            return PreviewArgumentAction

        def _get_preview_arg_message(self):
            # "Argument xxx"
            subject = "{} '{}'".format(self.object_type.capitalize(), self.target)
            return status_tag_messages['preview'].format(subject)

        options_list = kwargs.get('options_list', None)
        object_type = 'argument'

        if options_list is None:
            # convert argument dest
            target = '--{}'.format(argument_dest.replace('_', '-'))
        elif options_list:
            target = sorted(options_list, key=len)[-1]
        else:
            # positional argument
            target = kwargs.get('metavar', '<{}>'.format(argument_dest.upper()))
            object_type = 'positional argument'

        preview_info = PreviewItem(
            cli_ctx=self.command_loader.cli_ctx,
            target=target,
            object_type=object_type,
            message_func=_get_preview_arg_message
        )
        kwargs['preview_info'] = preview_info
        kwargs['action'] = _handle_argument_preview(preview_info)
        return kwargs

    def _handle_experimentals(self, argument_dest, **kwargs):

        if not kwargs.get('is_experimental', False):
            return kwargs

        def _handle_argument_experimental(experimental_info):

            parent_class = self._get_parent_class(**kwargs)

            class ExperimentalArgumentAction(parent_class):

                def __call__(self, parser, namespace, values, option_string=None):
                    if not hasattr(namespace, '_argument_experimentals'):
                        setattr(namespace, '_argument_experimentals', [experimental_info])
                    else:
                        namespace._argument_experimentals.append(experimental_info)  # pylint: disable=protected-access
                    try:
                        super().__call__(parser, namespace, values, option_string)
                    except NotImplementedError:
                        setattr(namespace, self.dest, values)

            return ExperimentalArgumentAction

        def _get_experimental_arg_message(self):
            # "Argument xxx"
            subject = "{} '{}'".format(self.object_type.capitalize(), self.target)
            return status_tag_messages['experimental'].format(subject)

        options_list = kwargs.get('options_list', None)
        object_type = 'argument'

        if options_list is None:
            # convert argument dest
            target = '--{}'.format(argument_dest.replace('_', '-'))
        elif options_list:
            target = sorted(options_list, key=len)[-1]
        else:
            # positional argument
            target = kwargs.get('metavar', '<{}>'.format(argument_dest.upper()))
            object_type = 'positional argument'

        experimental_info = ExperimentalItem(
            self.command_loader.cli_ctx,
            target=target,
            object_type=object_type,
            message_func=_get_experimental_arg_message
        )
        kwargs['experimental_info'] = experimental_info
        kwargs['action'] = _handle_argument_experimental(experimental_info)
        return kwargs

    # pylint: disable=inconsistent-return-statements
    def deprecate(self, **kwargs):

        def _get_deprecated_arg_message(self):
            msg = "{} '{}' has been deprecated and will be removed ".format(
                self.object_type, self.target).capitalize()
            if self.expiration:
                msg += "in version '{}'.".format(self.expiration)
            else:
                msg += 'in a future release.'
            if self.redirect:
                msg += " Use '{}' instead.".format(self.redirect)
            return msg

        self._check_stale()
        if not self._applicable():
            return

        target = kwargs.get('target', '')
        kwargs['object_type'] = 'option' if target.startswith('-') else 'argument'
        kwargs['message_func'] = _get_deprecated_arg_message
        return Deprecated(self.command_loader.cli_ctx, **kwargs)

    def argument(self, argument_dest, arg_type=None, **kwargs):
        """ Register an argument for the given command scope using a knack.arguments.CLIArgumentType

        :param argument_dest: The destination argument to add this argument type to
        :type argument_dest: str
        :param arg_type: Predefined CLIArgumentType definition to register, as modified by any provided kwargs.
        :type arg_type: knack.arguments.CLIArgumentType
        :param kwargs: Possible values: `options_list`, `validator`, `completer`, `nargs`, `action`, `const`, `default`,
                       `type`, `choices`, `required`, `help`, `metavar`, `is_preview`, `is_experimental`,
                       `deprecate_info`. See /docs/arguments.md.
        """
        self._check_stale()
        if not self._applicable():
            return

        deprecate_action = self._handle_deprecations(argument_dest, **kwargs)
        if deprecate_action:
            kwargs['action'] = deprecate_action

        is_preview = kwargs.get('is_preview', False)
        is_experimental = kwargs.get('is_experimental', False)

        if is_preview and is_experimental:
            from .commands import PREVIEW_EXPERIMENTAL_CONFLICT_ERROR
            raise CLIError(PREVIEW_EXPERIMENTAL_CONFLICT_ERROR.format('argument', argument_dest))

        kwargs = self._handle_previews(argument_dest, **kwargs)
        kwargs = self._handle_experimentals(argument_dest, **kwargs)

        self.command_loader.argument_registry.register_cli_argument(self.command_scope,
                                                                    argument_dest,
                                                                    arg_type,
                                                                    **kwargs)

    def positional(self, argument_dest, arg_type=None, **kwargs):
        """ Register a positional argument for the given command scope using a knack.arguments.CLIArgumentType

        :param argument_dest: The destination argument to add this argument type to
        :type argument_dest: str
        :param arg_type: Predefined CLIArgumentType definition to register, as modified by any provided kwargs.
        :type arg_type: knack.arguments.CLIArgumentType
        :param kwargs: Possible values: `validator`, `completer`, `nargs`, `action`, `const`, `default`,
                       `type`, `choices`, `required`, `help`, `metavar`, `is_preview`, `is_experimental`,
                       `deprecate_info`. See /docs/arguments.md.
        """
        self._check_stale()
        if not self._applicable():
            return

        if self.command_scope not in self.command_loader.command_table:
            raise ValueError("command authoring error: positional argument '{}' cannot be registered to a group-level "
                             "scope '{}'. It must be registered to a specific command.".format(
                                 argument_dest, self.command_scope))

        # Before adding the new positional arg, ensure that there are no existing positional arguments
        # registered for this command.
        command_args = self.command_loader.argument_registry.arguments[self.command_scope]
        positional_args = {k: v for k, v in command_args.items() if v.settings.get('options_list') == []}
        if positional_args and argument_dest not in positional_args:
            raise CLIError("command authoring error: commands may have, at most, one positional argument. '{}' already "
                           "has positional argument: {}.".format(self.command_scope, ' '.join(positional_args.keys())))

        kwargs['options_list'] = []

        deprecate_action = self._handle_deprecations(argument_dest, **kwargs)
        if deprecate_action:
            kwargs['action'] = deprecate_action

        kwargs = self._handle_previews(argument_dest, **kwargs)
        kwargs = self._handle_experimentals(argument_dest, **kwargs)

        self.command_loader.argument_registry.register_cli_argument(self.command_scope,
                                                                    argument_dest,
                                                                    arg_type,
                                                                    **kwargs)

    def ignore(self, argument_dest, **kwargs):
        """ Register an argument with type knack.arguments.ignore_type (hidden/ignored)

        :param argument_dest: The destination argument to apply the ignore type to
        :type argument_dest: str
        """
        self._check_stale()
        if not self._applicable():
            return

        dest_option = ['--__{}'.format(argument_dest.upper())]
        self.argument(argument_dest, arg_type=ignore_type, options_list=dest_option, **kwargs)

    def extra(self, argument_dest, **kwargs):
        """Register extra parameters for the given command. Typically used to augment auto-command built
        commands to add more parameters than the specific SDK method introspected.

        :param argument_dest: The destination argument to add this argument type to
        :type argument_dest: str
        :param kwargs: Possible values: `options_list`, `validator`, `completer`, `nargs`, `action`, `const`, `default`,
                       `type`, `choices`, `required`, `help`, `metavar`, `is_preview`, `is_experimental`,
                       `deprecate_info`. See /docs/arguments.md.
        """
        self._check_stale()
        if not self._applicable():
            return

        if self.command_scope in self.command_loader.command_group_table:
            raise ValueError("command authoring error: extra argument '{}' cannot be registered to a group-level "
                             "scope '{}'. It must be registered to a specific command.".format(
                                 argument_dest, self.command_scope))

        deprecate_action = self._handle_deprecations(argument_dest, **kwargs)
        if deprecate_action:
            kwargs['action'] = deprecate_action

        kwargs = self._handle_previews(argument_dest, **kwargs)
        kwargs = self._handle_experimentals(argument_dest, **kwargs)

        self.command_loader.extra_argument_registry[self.command_scope][argument_dest] = CLICommandArgument(
            argument_dest, **kwargs)


class IgnoreAction(argparse.Action):  # pylint: disable=too-few-public-methods
    """ Show the argument as unrecognized if it is called """

    def __call__(self, parser, namespace, values, option_string=None):
        raise argparse.ArgumentError(None, 'unrecognized argument: {} {}'.format(
            option_string, values or ''))


class CaseInsensitiveList(list):
    """ Determine if a choice is in a choice list in a case-insensitive manner """

    def __contains__(self, other):
        return next((True for x in self if other.lower() == x.lower()), False)


def enum_choice_list(data):
    """ Creates the argparse choices and type kwargs for a supplied enum type or list of strings """

    # transform enum types, otherwise assume list of string choices
    if not data:
        return {}
    try:
        choices = [x.value for x in data]
    except AttributeError:
        choices = data

    def _type(value):
        return next((x for x in choices if x.lower() == value.lower()), value) if value else value
    params = {
        'choices': CaseInsensitiveList(choices),
        'type': _type
    }
    return params

# GLOBAL ARGUMENT DEFINITIONS


ignore_type = CLIArgumentType(
    help=argparse.SUPPRESS,
    nargs='?',
    action=IgnoreAction,
    required=False)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys
import unittest
from unittest import mock
from io import StringIO

from knack.prompting import (verify_is_a_tty, NoTTYException, _INVALID_PASSWORD_MSG, prompt,
                             prompt_int, prompt_pass, prompt_y_n, prompt_t_f, prompt_choice_list)


@unittest.skipIf(sys.version_info[0] == 2, "Can't modify isatty in Python 2 as it's read-only")
class TestPrompting(unittest.TestCase):

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_tty_no_exception(self, _):
        verify_is_a_tty()

    @mock.patch('sys.stdin.isatty', return_value=False)
    def test_no_tty_should_raise_exception(self, _):
        with self.assertRaises(NoTTYException):
            verify_is_a_tty()

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_msg(self, _):
        expected_result = 'This is my response.'
        with mock.patch('knack.prompting._input', return_value=expected_result):
            actual_result = prompt('Please enter some text: ')
            self.assertEqual(expected_result, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_msg_empty_response(self, _):
        expected_result = ''
        with mock.patch('knack.prompting._input', return_value=expected_result):
            actual_result = prompt('Please enter some text: ')
            self.assertEqual(expected_result, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_msg_question_no_help_string(self, _):
        expected_result = '?'
        with mock.patch('knack.prompting._input', return_value='?'):
            actual_result = prompt('Please enter some text: ')
            self.assertEqual(expected_result, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_msg_question_with_help_string(self, _):
        expected_result = 'My response'
        with mock.patch('knack.prompting._input', side_effect=['?', expected_result]):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt('Please enter some text: ', help_string='Anything you want!')
                self.assertEqual(expected_result, actual_result)
                self.assertIn('Anything you want!', mock_stdout.getvalue())

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_int(self, _):
        my_response = '42'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_int('Please enter a number: ')
            self.assertEqual(int(my_response), actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_int_empty_response(self, _):
        my_response = ''
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('knack.prompting._input', side_effect=[my_response]):
                    prompt_int('Please enter some text: ')
            mock_log_warn.assert_called_once_with('%s is not a valid number', my_response)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_int_nan(self, _):
        my_response = 'This is clearly not a number.'
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('knack.prompting._input', side_effect=[my_response]):
                    prompt_int('Please enter some text: ')
            mock_log_warn.assert_called_once_with('%s is not a valid number', my_response)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_int_question_no_help_string(self, _):
        my_response = '?'
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('knack.prompting._input', side_effect=['?']):
                    prompt_int('Please enter a number: ')
            mock_log_warn.assert_called_once_with('%s is not a valid number', my_response)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_int_question_with_help_string(self, _):
        my_response = '42'
        with mock.patch('knack.prompting._input', side_effect=['?', my_response]):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt_int('Please enter a number: ', help_string='Anything you want!')
                self.assertEqual(int(my_response), actual_result)
                self.assertIn('Anything you want!', mock_stdout.getvalue())

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass(self, _):
        my_password = '7ndBkS3zKQazD5N3zzstubZq'
        with mock.patch('getpass.getpass', return_value=my_password):
            actual_result = prompt_pass()
            self.assertEqual(my_password, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_empty_response(self, _):
        my_password = ''
        with mock.patch('getpass.getpass', return_value=my_password):
            actual_result = prompt_pass()
            self.assertEqual(my_password, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_custom_msg(self, _):
        my_password = '7ndBkS3zKQazD5N3zzstubZq'
        with mock.patch('getpass.getpass', return_value=my_password):
            actual_result = prompt_pass(msg='A Custom password message: ')
            self.assertEqual(my_password, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_question_no_help_string(self, _):
        expected_result = '?'
        with mock.patch('getpass.getpass', return_value='?'):
            actual_result = prompt_pass()
            self.assertEqual(expected_result, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_question_with_help_string(self, _):
        my_password = '7ndBkS3zKQazD5N3zzstubZq'
        with mock.patch('getpass.getpass', side_effect=['?', my_password]):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt_pass(help_string='Anything you want!')
                self.assertEqual(my_password, actual_result)
                self.assertIn('Anything you want!', mock_stdout.getvalue())

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_confirm_valid(self, _):
        my_password = '7ndBkS3zKQazD5N3zzstubZq'
        with mock.patch('getpass.getpass', side_effect=[my_password, my_password]):
            actual_result = prompt_pass(confirm=True)
            self.assertEqual(my_password, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_confirm_invalid(self, _):
        my_password1 = '7ndBkS3zKQazD5N3zzstubZq'
        my_password2 = 'LTQ9haNMCSGp8p2uQHw2K9xf'
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('getpass.getpass', side_effect=[my_password1, my_password2]):
                    prompt_pass(confirm=True)
            mock_log_warn.assert_called_once_with(_INVALID_PASSWORD_MSG)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_pass_confirm_invalid_then_valid(self, _):
        my_password1 = '7ndBkS3zKQazD5N3zzstubZq'
        my_password2 = 'LTQ9haNMCSGp8p2uQHw2K9xf'
        with mock.patch('getpass.getpass', side_effect=[my_password1, my_password2, my_password2, my_password2]):
            with mock.patch('logging.Logger.warning') as mock_log_warn:
                actual_result = prompt_pass(confirm=True)
                mock_log_warn.assert_called_once_with(_INVALID_PASSWORD_MSG)
                self.assertEqual(my_password2, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_yes(self, _):
        my_response = 'y'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_y_n('Do you accept?')
            self.assertTrue(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_no(self, _):
        my_response = 'n'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_y_n('Do you accept?')
            self.assertFalse(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_yes_caps(self, _):
        my_response = 'Y'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_y_n('Do you accept?')
            self.assertTrue(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_no_caps(self, _):
        my_response = 'N'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_y_n('Do you accept?')
            self.assertFalse(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_empty_response(self, _):
        with self.assertRaises(StopIteration):
            with mock.patch('knack.prompting._input', side_effect=['']):
                prompt_y_n('Do you accept?')

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_question_no_help_string(self, _):
        with self.assertRaises(StopIteration):
            with mock.patch('knack.prompting._input', side_effect=['?']):
                prompt_y_n('Do you accept?')

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_question_with_help_string(self, _):
        with mock.patch('knack.prompting._input', side_effect=['?', 'y']):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt_y_n('Do you accept?', help_string='y to accept conditions; no otherwise')
                self.assertTrue(actual_result)
                self.assertIn('y to accept conditions; no otherwise', mock_stdout.getvalue())

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_y_n_default(self, _):
        with mock.patch('knack.prompting._input', return_value=''):
            actual_result = prompt_y_n('Do you accept?', default='y')
            self.assertTrue(actual_result)

# HERE
    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_yes(self, _):
        my_response = 't'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_t_f('Do you accept?')
            self.assertTrue(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_no(self, _):
        my_response = 'f'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_t_f('Do you accept?')
            self.assertFalse(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_yes_caps(self, _):
        my_response = 'T'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_t_f('Do you accept?')
            self.assertTrue(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_no_caps(self, _):
        my_response = 'F'
        with mock.patch('knack.prompting._input', return_value=my_response):
            actual_result = prompt_t_f('Do you accept?')
            self.assertFalse(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_empty_response(self, _):
        with self.assertRaises(StopIteration):
            with mock.patch('knack.prompting._input', side_effect=['']):
                prompt_t_f('Do you accept?')

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_question_no_help_string(self, _):
        with self.assertRaises(StopIteration):
            with mock.patch('knack.prompting._input', side_effect=['?']):
                prompt_t_f('Do you accept?')

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_question_with_help_string(self, _):
        with mock.patch('knack.prompting._input', side_effect=['?', 't']):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt_t_f('Do you accept?', help_string='t to accept conditions; no otherwise')
                self.assertTrue(actual_result)
                self.assertIn('t to accept conditions; no otherwise', mock_stdout.getvalue())

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_t_f_default(self, _):
        with mock.patch('knack.prompting._input', return_value=''):
            actual_result = prompt_t_f('Do you accept?', default='t')
            self.assertTrue(actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_choice_list(self, _):
        a_list = ['red', 'blue', 'yellow', 'green']
        with mock.patch('knack.prompting._input', return_value='3'):
            actual_result = prompt_choice_list('What is your favourite color?', a_list)
            self.assertEqual(2, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_choice_list_with_name_desc(self, _):
        a_list = [{'name': 'red', 'desc': ' A desc.'},
                  {'name': 'blue', 'desc': ' A desc.'},
                  {'name': 'yellow', 'desc': ' A desc.'},
                  {'name': 'green', 'desc': ' A desc.'}]
        with mock.patch('knack.prompting._input', return_value='2'):
            actual_result = prompt_choice_list('What is your favourite color?', a_list)
            self.assertEqual(1, actual_result)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_choice_list_invalid_choice(self, _):
        a_list = ['red', 'blue', 'yellow', 'green']
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('knack.prompting._input', side_effect=['5']):
                    prompt_choice_list('What is your favourite color?', a_list)
            mock_log_warn.assert_called_once_with('Valid values are %s', mock.ANY)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_choice_list_question_no_help_string(self, _):
        a_list = ['red', 'blue', 'yellow', 'green']
        with mock.patch('logging.Logger.warning') as mock_log_warn:
            with self.assertRaises(StopIteration):
                with mock.patch('knack.prompting._input', side_effect=['?']):
                    prompt_choice_list('What is your favourite color?', a_list)
            mock_log_warn.assert_called_once_with('Valid values are %s', mock.ANY)

    @mock.patch('sys.stdin.isatty', return_value=True)
    def test_prompt_choice_list_question_with_help_string(self, _):
        a_list = ['red', 'blue', 'yellow', 'green']
        with mock.patch('knack.prompting._input', side_effect=['?', '1']):
            with mock.patch('sys.stdout', new_callable=StringIO) as mock_stdout:
                actual_result = prompt_choice_list('What is your favourite color?',
                                                   a_list,
                                                   help_string='Your real favourite.')
                self.assertEqual(0, actual_result)
            self.assertIn('Your real favourite.', mock_stdout.getvalue())



if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys
from collections import defaultdict

from .commands import CLICommandsLoader
from .deprecation import ImplicitDeprecated, resolve_deprecate_info
from .events import (EVENT_INVOKER_PRE_CMD_TBL_CREATE, EVENT_INVOKER_POST_CMD_TBL_CREATE,
                     EVENT_INVOKER_CMD_TBL_LOADED, EVENT_INVOKER_PRE_PARSE_ARGS,
                     EVENT_INVOKER_POST_PARSE_ARGS, EVENT_INVOKER_TRANSFORM_RESULT,
                     EVENT_INVOKER_FILTER_RESULT)
from .experimental import ImplicitExperimentalItem, resolve_experimental_info
from .help import CLIHelp
from .log import CLILogging
from .parser import CLICommandParser
from .preview import ImplicitPreviewItem, resolve_preview_info
from .util import CLIError, CtxTypeError, CommandResultItem, todict


class CommandInvoker(object):

    def __init__(self,
                 cli_ctx=None,
                 parser_cls=CLICommandParser,
                 commands_loader_cls=CLICommandsLoader,
                 help_cls=CLIHelp,
                 initial_data=None):
        """ Manages a single invocation of the CLI (i.e. running a command)

        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        :param parser_cls: A class to handle command parsing
        :type parser_cls: knack.parser.CLICommandParser
        :param commands_loader_cls: A class to handle loading commands
        :type commands_loader_cls: knack.commands.CLICommandsLoader
        :param help_cls: A class to handle help
        :type help_cls: knack.help.CLIHelp
        :param initial_data: The initial in-memory collection for this command invocation
        :type initial_data: dict
        """
        from .cli import CLI
        if cli_ctx is not None and not isinstance(cli_ctx, CLI):
            raise CtxTypeError(cli_ctx)
        self.cli_ctx = cli_ctx
        # In memory collection of key-value data for this current invocation This does not persist between invocations.
        self.data = initial_data or defaultdict(lambda: None)
        self.data['command'] = 'unknown'
        self._global_parser = parser_cls.create_global_parser(cli_ctx=self.cli_ctx)
        self.help = help_cls(cli_ctx=self.cli_ctx)
        self.parser = parser_cls(cli_ctx=self.cli_ctx, cli_help=self.help,
                                 prog=self.cli_ctx.name, parents=[self._global_parser])
        self.commands_loader = commands_loader_cls(cli_ctx=self.cli_ctx)

    def _filter_params(self, args):
        # Consider - we are using any args that start with an underscore (_) as 'private'
        # arguments and remove them from the arguments that we pass to the actual function.
        params = {key: value
                  for key, value in args.__dict__.items()
                  if not key.startswith('_')}
        params.pop('func', None)
        params.pop('command', None)
        return params

    def _rudimentary_get_command(self, args):
        """ Rudimentary parsing to get the command """
        nouns = []
        command_names = self.commands_loader.command_table.keys()
        for arg in args:
            if arg and arg[0] != '-':
                nouns.append(arg)
            else:
                break

        def _find_args(args):
            search = ' '.join(args).lower()
            return next((x for x in command_names if x.startswith(search)), False)

        # since the command name may be immediately followed by a positional arg, strip those off
        while nouns and not _find_args(nouns):
            del nouns[-1]

        # ensure the command string is case-insensitive
        for i in range(len(nouns)):
            args[i] = args[i].lower()

        return ' '.join(nouns)

    def _validate_cmd_level(self, ns, cmd_validator):
        if cmd_validator:
            cmd_validator(ns)
        try:
            delattr(ns, '_command_validator')
        except AttributeError:
            pass

    def _validate_arg_level(self, ns, **_):
        for validator in getattr(ns, '_argument_validators', []):
            validator(ns)
        try:
            delattr(ns, '_argument_validators')
        except AttributeError:
            pass

    def _validation(self, parsed_ns):
        try:
            cmd_validator = getattr(parsed_ns, '_command_validator', None)
            if cmd_validator:
                self._validate_cmd_level(parsed_ns, cmd_validator)
            else:
                self._validate_arg_level(parsed_ns)
        except CLIError:
            raise
        except Exception:  # pylint: disable=broad-except
            err = sys.exc_info()[1]
            getattr(parsed_ns, '_parser', self.parser).validation_error(str(err))

    # pylint: disable=too-many-statements
    def execute(self, args):
        """ Executes the command invocation

        :param args: The command arguments for this invocation
        :type args: list
        :return: The command result
        :rtype: knack.util.CommandResultItem
        """

        self.cli_ctx.raise_event(EVENT_INVOKER_PRE_CMD_TBL_CREATE, args=args)
        cmd_tbl = self.commands_loader.load_command_table(args)
        command = self._rudimentary_get_command(args)
        self.cli_ctx.invocation.data['command_string'] = command
        self.commands_loader.load_arguments(command)

        self.cli_ctx.raise_event(EVENT_INVOKER_POST_CMD_TBL_CREATE, cmd_tbl=cmd_tbl)
        self.parser.load_command_table(self.commands_loader)
        self.cli_ctx.raise_event(EVENT_INVOKER_CMD_TBL_LOADED, parser=self.parser)

        arg_check = [a for a in args if a not in
                     (CLILogging.DEBUG_FLAG, CLILogging.VERBOSE_FLAG, CLILogging.ONLY_SHOW_ERRORS_FLAG)]
        if not arg_check:
            self.cli_ctx.completion.enable_autocomplete(self.parser)
            subparser = self.parser.subparsers[tuple()]
            self.help.show_welcome(subparser)
            return CommandResultItem(None, exit_code=0)

        if args[0].lower() == 'help':
            args[0] = '--help'

        self.cli_ctx.completion.enable_autocomplete(self.parser)

        self.cli_ctx.raise_event(EVENT_INVOKER_PRE_PARSE_ARGS, args=args)
        parsed_args = self.parser.parse_args(args)
        self.cli_ctx.raise_event(EVENT_INVOKER_POST_PARSE_ARGS, command=parsed_args.command, args=parsed_args)

        self._validation(parsed_args)

        # save the command name (leaf in the tree)
        self.data['command'] = parsed_args.command
        cmd = parsed_args.func
        if hasattr(parsed_args, 'cmd'):
            parsed_args.cmd = cmd
        deprecations = getattr(parsed_args, '_argument_deprecations', [])
        if cmd.deprecate_info:
            deprecations.append(cmd.deprecate_info)

        previews = getattr(parsed_args, '_argument_previews', [])
        if cmd.preview_info:
            previews.append(cmd.preview_info)

        experimentals = getattr(parsed_args, '_argument_experimentals', [])
        if cmd.experimental_info:
            experimentals.append(cmd.experimental_info)

        params = self._filter_params(parsed_args)

        # search for implicit deprecation
        path_comps = cmd.name.split()[:-1]
        implicit_deprecate_info = None
        while path_comps and not implicit_deprecate_info:
            implicit_deprecate_info = resolve_deprecate_info(self.cli_ctx, ' '.join(path_comps))
            del path_comps[-1]

        if implicit_deprecate_info:
            deprecate_kwargs = implicit_deprecate_info.__dict__.copy()
            deprecate_kwargs['object_type'] = 'command'
            del deprecate_kwargs['_get_tag']
            del deprecate_kwargs['_get_message']
            deprecations.append(ImplicitDeprecated(cli_ctx=self.cli_ctx, **deprecate_kwargs))

        # search for implicit preview
        path_comps = cmd.name.split()[:-1]
        implicit_preview_info = None
        while path_comps and not implicit_preview_info:
            implicit_preview_info = resolve_preview_info(self.cli_ctx, ' '.join(path_comps))
            del path_comps[-1]

        if implicit_preview_info:
            preview_kwargs = implicit_preview_info.__dict__.copy()
            preview_kwargs['object_type'] = 'command'
            previews.append(ImplicitPreviewItem(cli_ctx=self.cli_ctx, **preview_kwargs))

        # search for implicit experimental
        path_comps = cmd.name.split()[:-1]
        implicit_experimental_info = None
        while path_comps and not implicit_experimental_info:
            implicit_experimental_info = resolve_experimental_info(self.cli_ctx, ' '.join(path_comps))
            del path_comps[-1]

        if implicit_experimental_info:
            experimental_kwargs = implicit_experimental_info.__dict__.copy()
            experimental_kwargs['object_type'] = 'command'
            experimentals.append(ImplicitExperimentalItem(cli_ctx=self.cli_ctx, **experimental_kwargs))

        if not self.cli_ctx.only_show_errors:
            for d in deprecations:
                print(d.message, file=sys.stderr)
            for p in previews:
                print(p.message, file=sys.stderr)
            for p in experimentals:
                print(p.message, file=sys.stderr)

        cmd_result = parsed_args.func(params)
        cmd_result = todict(cmd_result)

        event_data = {'result': cmd_result}
        self.cli_ctx.raise_event(EVENT_INVOKER_TRANSFORM_RESULT, event_data=event_data)
        self.cli_ctx.raise_event(EVENT_INVOKER_FILTER_RESULT, event_data=event_data)

        return CommandResultItem(event_data['result'],
                                 exit_code=0,
                                 table_transformer=cmd_tbl[parsed_args.command].table_transformer,
                                 is_query_active=self.data['query_active'],
                                 raw_result=cmd_result)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from .base import ScenarioTest, LiveTest
from .exceptions import CliTestError
from .checkers import (JMESPathCheck, JMESPathCheckExists, JMESPathCheckGreaterThan, NoneCheck,
                       StringCheck, StringContainCheck)
from .decorators import live_only, record_only

__all__ = ['ScenarioTest', 'LiveTest', 'CliTestError', 'JMESPathCheck', 'JMESPathCheckExists', 'NoneCheck',
           'live_only', 'record_only', 'StringCheck', 'StringContainCheck',
           'JMESPathCheckGreaterThan']



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from unittest import mock
import logging
import os
import re
import shutil
import sys
import tempfile
from io import StringIO

from knack.cli import CLI, CLICommandsLoader, CommandInvoker
from knack.log import CLI_LOGGER_NAME

TEMP_FOLDER_NAME = "knack_temp"


def redirect_io(func):

    original_stderr = sys.stderr
    original_stdout = sys.stdout

    def wrapper(self):
        # Ensure a clean startup - no log handlers
        root_logger = logging.getLogger()
        cli_logger = logging.getLogger(CLI_LOGGER_NAME)
        root_logger.handlers.clear()
        cli_logger.handlers.clear()

        sys.stdout = sys.stderr = self.io = StringIO()
        func(self)
        self.io.close()
        sys.stdout = original_stdout
        sys.stderr = original_stderr

        # Remove the handlers added by CLI, so that the next invoke call init them again with the new stderr
        # Otherwise, the handlers will write to a closed StringIO from a preview test
        root_logger.handlers.clear()
        cli_logger.handlers.clear()
    return wrapper


def disable_color(func):
    def wrapper(self):
        self.cli_ctx.enable_color = False
        func(self)
        self.cli_ctx.enable_color = True
    return wrapper


def _remove_control_sequence(string):
    return re.sub(r'\x1b[^m]+m', '', string)


def _remove_whitespace(string):
    return re.sub(r'\s', '', string)


def assert_in_multi_line(sub_string, string):
    # assert sub_string is in string, with all whitespaces, line breaks and control sequences ignored
    assert _remove_whitespace(sub_string) in _remove_control_sequence(_remove_whitespace(string))


class MockContext(CLI):

    def __init__(self):
        super().__init__(config_dir=new_temp_folder())
        loader = CLICommandsLoader(cli_ctx=self)
        invocation = mock.MagicMock(spec=CommandInvoker)
        invocation.data = {}
        setattr(self, 'commands_loader', loader)
        setattr(self, 'invocation', invocation)


class DummyCLI(CLI):

    def get_cli_version(self):
        return '0.1.0'

    def __init__(self, **kwargs):
        kwargs['config_dir'] = new_temp_folder()
        super().__init__(**kwargs)
        # Force to enable color
        self.enable_color = True


def new_temp_folder():
    temp_dir = os.path.join(tempfile.gettempdir(), TEMP_FOLDER_NAME)
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)
    return temp_dir



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys
import unittest

from knack.commands import CLICommandsLoader, CommandGroup
from knack.arguments import CLIArgumentType, CLICommandArgument, ArgumentsContext
from tests.util import MockContext


def _dictContainsSubset(expected, actual):
    """Checks whether actual is a superset of expected.
       Helper for deprecated assertDictContainsSubset"""
    missing = False
    mismatched = False
    for key, value in expected.items():
        if key not in actual:
            missing = True
        elif value != actual[key]:
            mismatched = True
    return False if missing or mismatched else True


class TestCommandRegistration(unittest.TestCase):

    def setUp(self):
        self.mock_ctx = MockContext()

    @staticmethod
    def sample_command_handler(group_name, resource_name, opt_param=None, expand=None):
        """
        The operation to get a virtual machine.

        :param group_name: The name of the group.
        :type group_name: str
        :param resource_name: The name of the resource.
        :type resource_name: str
        :param opt_param: Used to verify reflection correctly
        identifies optional params.
        :type opt_param: object
        :param expand: The expand expression to apply on the operation.
        :type expand: str
        :param dict custom_headers: headers that will be added to the request
        :param boolean raw: returns the direct response alongside the
            deserialized response
        """
        pass

    @staticmethod
    def sample_command_handler2(group_name, resource_name, opt_param=None, expand=None, custom_headers=None,
                                raw=False, **operation_config):
        pass

    def _set_command_name(self, command):
        self.mock_ctx.invocation.data['command_string'] = command
        return command

    def test_register_cli_argument(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test register sample-command')
        with CommandGroup(cl, 'test register', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__))
        with ArgumentsContext(cl, command_name) as ac:
            ac.argument('resource_name', CLIArgumentType(
                options_list=('--wonky-name', '-n'), metavar='RNAME', help='Completely WONKY name...',
                required=False
            ))
        cl.load_arguments(command_name)
        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 4, 'We expected exactly 4 arguments')
        some_expected_arguments = {
            'group_name': CLIArgumentType(dest='group_name', required=True),
            'resource_name': CLIArgumentType(dest='resource_name', required=False),
        }
        for probe in some_expected_arguments:
            existing = next(arg for arg in command_metadata.arguments if arg == probe)
            contains_subset = _dictContainsSubset(some_expected_arguments[existing].settings,
                                                  command_metadata.arguments[existing].options)
            self.assertTrue(contains_subset)
        self.assertEqual(command_metadata.arguments['resource_name'].options_list, ('--wonky-name', '-n'))

    def test_register_command_custom_excluded_params(self):
        command_name = self._set_command_name('test sample-command')
        ep = ['self', 'raw', 'custom_headers', 'operation_config', 'content_version', 'kwargs', 'client']
        cl = CLICommandsLoader(self.mock_ctx, excluded_command_handler_args=ep)
        with CommandGroup(cl, 'test', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler2.__name__))
        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        cl.load_arguments(command_name)
        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 4, 'We expected exactly 4 arguments')
        self.assertIn(command_name, cl.command_table)

    def test_register_command(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test register sample-command')
        with CommandGroup(cl, 'test register', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__))

        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        cl.load_arguments(command_name)
        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 4, 'We expected exactly 4 arguments')
        some_expected_arguments = {
            'group_name': CLIArgumentType(dest='group_name',
                                          required=True,
                                          help='The name of the group.'),
            'resource_name': CLIArgumentType(dest='resource_name',
                                             required=True,
                                             help='The name of the resource.'),
            'opt_param': CLIArgumentType(required=False,
                                         help='Used to verify reflection correctly identifies optional params.'),
            'expand': CLIArgumentType(required=False,
                                      help='The expand expression to apply on the operation.')
        }

        for probe in some_expected_arguments:
            existing = next(arg for arg in command_metadata.arguments if arg == probe)
            contains_subset = _dictContainsSubset(some_expected_arguments[existing].settings,
                                                  command_metadata.arguments[existing].options)
            self.assertTrue(contains_subset)
        self.assertEqual(command_metadata.arguments['resource_name'].options_list, ['--resource-name'])

    def test_register_command_group_with_no_group_name(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('sample-command')
        with CommandGroup(cl, None, '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__))

        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        self.assertIn(command_name, cl.command_table)

    def test_register_command_confirmation_bool(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test sample-command')
        with CommandGroup(cl, 'test', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__),
                      confirmation=True)
        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        cl.load_arguments(command_name)
        command_metadata = cl.command_table[command_name]
        self.assertIn('yes', command_metadata.arguments)
        self.assertEqual(command_metadata.arguments['yes'].type.settings['action'], 'store_true')
        self.assertIs(command_metadata.confirmation, True)

    def test_register_command_confirmation_callable(self):
        cl = CLICommandsLoader(self.mock_ctx)

        def confirm_callable(_):
            pass
        command_name = self._set_command_name('test sample-command')
        with CommandGroup(cl, 'test', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__),
                      confirmation=confirm_callable)
        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        cl.load_arguments(command_name)
        command_metadata = cl.command_table[command_name]
        self.assertIn('yes', command_metadata.arguments)
        self.assertEqual(command_metadata.arguments['yes'].type.settings['action'], 'store_true')
        self.assertIs(command_metadata.confirmation, confirm_callable)

    def test_register_cli_argument_with_overrides(self):
        cl = CLICommandsLoader(self.mock_ctx)
        base_type = CLIArgumentType(options_list=['--foo', '-f'], metavar='FOO', help='help1')
        derived_type = CLIArgumentType(base_type=base_type, help='help2')
        with CommandGroup(cl, 'test', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-get', '{}.{}'.format(TestCommandRegistration.__name__,
                                                   TestCommandRegistration.sample_command_handler.__name__))
            g.command('command sample-get-1', '{}.{}'.format(TestCommandRegistration.__name__,
                                                             TestCommandRegistration.sample_command_handler.__name__))
            g.command('command sample-get-2', '{}.{}'.format(TestCommandRegistration.__name__,
                                                             TestCommandRegistration.sample_command_handler.__name__))
        self.assertEqual(len(cl.command_table), 3, 'We expect exactly three commands in the command table')

        def test_with_command(command, target_value):
            self._set_command_name(command)
            with ArgumentsContext(cl, 'test') as c:
                c.argument('resource_name', base_type)
            with ArgumentsContext(cl, 'test command') as c:
                c.argument('resource_name', derived_type)
            with ArgumentsContext(cl, 'test command sample-get-2') as c:
                c.argument('resource_name', derived_type, help='help3')
            cl.load_arguments(command)
            command1 = cl.command_table[command].arguments['resource_name']
            self.assertEqual(command1.options['help'], target_value)

        test_with_command('test sample-get', 'help1')
        test_with_command('test command sample-get-1', 'help2')
        test_with_command('test command sample-get-2', 'help3')

    def test_register_extra_cli_argument(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test register sample-command')
        with CommandGroup(cl, 'test register', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__))
        with ArgumentsContext(cl, command_name) as ac:
            ac.extra('added_param', options_list=('--added-param',),
                     metavar='ADDED', help='Just added this right now!', required=True)
        cl.load_arguments(command_name)
        self.assertEqual(len(cl.command_table), 1, 'We expect exactly one command in the command table')
        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 5, 'We expected exactly 5 arguments')

        some_expected_arguments = {
            'added_param': CLIArgumentType(dest='added_param', required=True)
        }

        for probe in some_expected_arguments:
            existing = next(arg for arg in command_metadata.arguments if arg == probe)
            contains_subset = _dictContainsSubset(some_expected_arguments[existing].settings,
                                                  command_metadata.arguments[existing].options)
            self.assertTrue(contains_subset)

    def test_register_ignore_cli_argument(self):
        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test register sample-command')
        self.mock_ctx.invocation.data['command_string'] = command_name
        with CommandGroup(cl, 'test register', '{}#{{}}'.format(__name__)) as g:
            g.command('sample-command', '{}.{}'.format(TestCommandRegistration.__name__,
                                                       TestCommandRegistration.sample_command_handler.__name__))
        with ArgumentsContext(cl, 'test register') as ac:
            ac.argument('resource_name', options_list=['--this'])
        with ArgumentsContext(cl, 'test register sample-command') as ac:
            ac.ignore('resource_name')
            ac.argument('opt_param', options_list=['--this'])
        cl.load_arguments(command_name)
        self.assertNotEqual(cl.command_table[command_name].arguments['resource_name'].options_list,
                            cl.command_table[command_name].arguments['opt_param'].options_list,
                            "Name conflict in options list")

    def test_command_build_argument_help_text(self):
        def sample_sdk_method_with_weird_docstring(param_a, param_b, param_c):  # pylint: disable=unused-argument
            """
            An operation with nothing good.

            :param dict param_a:
            :param param_b: The name
            of
            nothing.
            :param param_c: The name
            of

            nothing2.
            """
            pass

        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('test command foo')
        setattr(sys.modules[__name__], sample_sdk_method_with_weird_docstring.__name__,
                sample_sdk_method_with_weird_docstring)
        with CommandGroup(cl, 'test command', '{}#{{}}'.format(__name__)) as g:
            g.command('foo', sample_sdk_method_with_weird_docstring.__name__)
        cl.load_arguments(command_name)
        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 3, 'We expected exactly 3 arguments')
        some_expected_arguments = {
            'param_a': CLIArgumentType(dest='param_a', required=True, help=''),
            'param_b': CLIArgumentType(dest='param_b', required=True, help='The name of nothing.'),
            'param_c': CLIArgumentType(dest='param_c', required=True, help='The name of nothing2.')
        }

        for probe in some_expected_arguments:
            existing = next(arg for arg in command_metadata.arguments if arg == probe)
            contains_subset = _dictContainsSubset(some_expected_arguments[existing].settings,
                                                  command_metadata.arguments[existing].options)
            self.assertTrue(contains_subset)

    def test_override_existing_option_string(self):
        arg = CLIArgumentType(options_list=('--funky', '-f'))
        updated_options_list = ('--something-else', '-s')
        arg.update(options_list=updated_options_list, validator=lambda: (), completer=lambda: ())
        self.assertEqual(arg.settings['options_list'], updated_options_list)
        self.assertIsNotNone(arg.settings['validator'])
        self.assertIsNotNone(arg.settings['completer'])

    def test_dont_override_existing_option_string(self):
        existing_options_list = ('--something-else', '-s')
        arg = CLIArgumentType(options_list=existing_options_list)
        arg.update()
        self.assertEqual(arg.settings['options_list'], existing_options_list)

    def test_override_remove_validator(self):
        existing_options_list = ('--something-else', '-s')
        arg = CLIArgumentType(options_list=existing_options_list,
                              validator=lambda *args, **kwargs: ())
        arg.update(validator=None)
        self.assertIsNone(arg.settings['validator'])

    def test_override_using_register_cli_argument(self):
        def sample_sdk_method(param_a):  # pylint: disable=unused-argument
            pass

        def test_validator_completer():
            pass

        cl = CLICommandsLoader(self.mock_ctx)
        command_name = self._set_command_name('override_using_register_cli_argument foo')
        setattr(sys.modules[__name__], sample_sdk_method.__name__, sample_sdk_method)
        with CommandGroup(cl, 'override_using_register_cli_argument', '{}#{{}}'.format(__name__)) as g:
            g.command('foo', sample_sdk_method.__name__)
        with ArgumentsContext(cl, 'override_using_register_cli_argument') as ac:
            ac.argument('param_a',
                        options_list=('--overridden', '-r'),
                        validator=test_validator_completer,
                        completer=test_validator_completer,
                        required=False)
        cl.load_arguments(command_name)

        command_metadata = cl.command_table[command_name]
        self.assertEqual(len(command_metadata.arguments), 1, 'We expected exactly 1 arguments')

        actual_arg = command_metadata.arguments['param_a']
        self.assertEqual(actual_arg.options_list, ('--overridden', '-r'))
        self.assertEqual(actual_arg.validator, test_validator_completer)
        self.assertEqual(actual_arg.completer, test_validator_completer)
        self.assertFalse(actual_arg.options['required'])

    def test_override_argtype_with_argtype(self):
        existing_options_list = ('--default', '-d')
        arg = CLIArgumentType(options_list=existing_options_list, validator=None, completer='base',
                              help='base', required=True)
        overriding_argtype = CLIArgumentType(options_list=('--overridden',), validator='overridden',
                                             completer=None, overrides=arg, help='overridden',
                                             required=CLIArgumentType.REMOVE)
        self.assertEqual(overriding_argtype.settings['validator'], 'overridden')
        self.assertIsNone(overriding_argtype.settings['completer'])
        self.assertEqual(overriding_argtype.settings['options_list'], ('--overridden',))
        self.assertEqual(overriding_argtype.settings['help'], 'overridden')
        self.assertEqual(overriding_argtype.settings['required'], CLIArgumentType.REMOVE)

        cmd_arg = CLICommandArgument(dest='whatever', argtype=overriding_argtype,
                                     help=CLIArgumentType.REMOVE)
        self.assertNotIn('required', cmd_arg.options)
        self.assertNotIn('help', cmd_arg.options)

    def test_cli_ctx_type_error(self):
        with self.assertRaises(TypeError):
            CLICommandsLoader(cli_ctx=object())


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import os
import logging
import sys
import unittest
from unittest import mock

from knack.arguments import ArgumentsContext
from knack.commands import CLICommandsLoader, CommandGroup
from tests.util import DummyCLI, redirect_io


# a dummy callback for arg-parse
def load_params(_):
    pass


def list_foo(my_param):
    print(str(my_param), end='')


class TestCommandWithConfiguredDefaults(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Ensure initialization has occurred correctly
        logging.basicConfig(level=logging.DEBUG)

    @classmethod
    def tearDownClass(cls):
        logging.shutdown()

    def _set_up_command_table(self, required):

        class TestCommandsLoader(CLICommandsLoader):

            def load_command_table(self, args):
                super().load_command_table(args)
                with CommandGroup(self, 'foo', '{}#{{}}'.format(__name__)) as g:
                    g.command('list', 'list_foo')
                return self.command_table

            def load_arguments(self, command):
                with ArgumentsContext(self, 'foo') as c:
                    c.argument('my_param', options_list='--my-param',
                               configured_default='param', required=required)
                super().load_arguments(command)
        self.cli_ctx = DummyCLI(commands_loader_cls=TestCommandsLoader)

    @mock.patch.dict(os.environ, {'CLI_DEFAULTS_PARAM': 'myVal'})
    @redirect_io
    def test_apply_configured_defaults_on_required_arg(self):
        self._set_up_command_table(required=True)
        self.cli_ctx.invoke('foo list'.split())
        actual = self.io.getvalue()
        expected = 'myVal'
        self.assertEqual(expected, actual)

    @redirect_io
    def test_no_configured_default_on_required_arg(self):
        self._set_up_command_table(required=True)
        with self.assertRaises(SystemExit):
            self.cli_ctx.invoke('foo list'.split())
        actual = self.io.getvalue()
        expected = 'required: --my-param'
        if sys.version_info[0] == 2:
            expected = 'argument --my-param is required'
        self.assertEqual(expected in actual, True)

    @mock.patch.dict(os.environ, {'CLI_DEFAULTS_PARAM': 'myVal'})
    @redirect_io
    def test_apply_configured_defaults_on_optional_arg(self):
        self._set_up_command_table(required=False)
        self.cli_ctx.invoke('foo list'.split())
        actual = self.io.getvalue()
        expected = 'myVal'
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# Verify that all *.py files have a license header in the file.

import os
import sys

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))

PY_LICENSE_HEADER = """# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""

env_folders = [
    os.path.join(ROOT_DIR, 'env'),
    os.path.join(ROOT_DIR, 'env27'),
    os.path.join(ROOT_DIR, '.tox'),
]


def contains_header(text):
    return PY_LICENSE_HEADER in text


def get_files_without_header():
    files_without_header = []
    for current_dir, _, files in os.walk(ROOT_DIR):
        # skip folders generated by virtual env
        if any(d for d in env_folders if d in current_dir):
            continue

        for a_file in files:
            if a_file.endswith('.py'):
                cur_file_path = os.path.join(current_dir, a_file)
                with open(cur_file_path, 'r', encoding='utf-8') as f:
                    file_text = f.read()

                if len(file_text) > 0 and not contains_header(file_text):
                    files_without_header.append((cur_file_path, file_text))
    return files_without_header


files_without_header = [file_path for file_path, file_contents in get_files_without_header()]

if files_without_header:
    print("Error: The following files don't have the required license headers:", file=sys.stderr)
    print('\n'.join(files_without_header), file=sys.stderr)
    print("Error: {} file(s) found without license headers.".format(len(files_without_header)), file=sys.stderr)
    sys.exit(1)
else:
    pass



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys
from setuptools import setup

VERSION = '0.14.0'

DEPENDENCIES = [
    'argcomplete',
    'jmespath',
    'packaging',
    'pygments',
    'pyyaml',
    'tabulate'
]

with open('README.rst', 'r') as f:
    README = f.read()

setup(
    name='knack',
    version=VERSION,
    description='A Command-Line Interface framework',
    long_description=README,
    license='MIT',
    author='Microsoft Corporation',
    author_email='azpycli@microsoft.com',
    url='https://github.com/microsoft/knack',
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'License :: OSI Approved :: MIT License',
    ],
    packages=['knack', 'knack.testsdk'],
    install_requires=DEPENDENCIES
)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


# commands should add entries to helps in the form: "group command": "YAML help"
helps = {}


def _load_help_file(delimiters):
    import yaml
    if delimiters in helps:
        return yaml.safe_load(helps[delimiters])
    return None



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import argparse

from .deprecation import Deprecated
from .events import EVENT_PARSER_GLOBAL_CREATE
from .log import get_logger
from .util import CtxTypeError

logger = get_logger(__name__)

# List of keyword arguments supported in argparse
# from https://github.com/python/cpython/blob/master/Lib/argparse.py#L748
ARGPARSE_SUPPORTED_KWARGS = [
    'option_strings',
    'dest',
    'nargs',
    'const',
    'default',
    'type',
    'choices',
    'required',
    'help',
    'metavar',
    'action',
    'default_value_source'
]


class CLICommandParser(argparse.ArgumentParser):

    @staticmethod
    def _sanitize_help_for_argparse(help_text):
        """Escape literal '%' for argparse unless the author already escaped.

        If a help string already contains ``%(...)`` placeholders or ``%%``,
        keep it as-is and rely on argparse's native formatting behavior.
        Otherwise, escape ``%`` so literal percent tokens don't break help
        processing in newer Python versions.
        """
        if not isinstance(help_text, str) or '%' not in help_text:
            return help_text
        if '%(' in help_text or '%%' in help_text:
            return help_text
        return help_text.replace('%', '%%')

    @staticmethod
    def create_global_parser(cli_ctx=None):
        global_parser = argparse.ArgumentParser(prog=cli_ctx.name, add_help=False)
        arg_group = global_parser.add_argument_group('global', 'Global Arguments')
        cli_ctx.raise_event(EVENT_PARSER_GLOBAL_CREATE, arg_group=arg_group)
        return global_parser

    @staticmethod
    def _add_argument(obj, arg):
        """ Only pass valid argparse kwargs to argparse.ArgumentParser.add_argument """
        argparse_options = {name: value for name, value in arg.options.items() if name in ARGPARSE_SUPPORTED_KWARGS}
        if 'help' in argparse_options:
            argparse_options['help'] = CLICommandParser._sanitize_help_for_argparse(argparse_options['help'])
        if arg.options_list:
            scrubbed_options_list = []
            for item in arg.options_list:

                if isinstance(item, Deprecated):
                    # don't add expired options to the parser
                    if item.expired():
                        continue

                    class _DeprecatedOption(str):
                        def __new__(cls, *args, **kwargs):
                            instance = str.__new__(cls, *args, **kwargs)
                            return instance

                    option = _DeprecatedOption(item.target)
                    setattr(option, 'deprecate_info', item)
                    item = option
                scrubbed_options_list.append(item)
            return obj.add_argument(*scrubbed_options_list, **argparse_options)

        if 'required' in argparse_options:
            del argparse_options['required']
        if 'metavar' not in argparse_options:
            argparse_options['metavar'] = '<{}>'.format(argparse_options['dest'].upper())
        return obj.add_argument(**argparse_options)

    @staticmethod
    def _expand_prefixed_files(args):
        """ Load arguments prefixed with '@' from file as string

        :param args: Arguments passed from command line
        :type args: list
        """
        for arg, _ in enumerate(args):
            if args[arg].startswith('@'):
                try:
                    logger.debug('Attempting to read file %s', args[arg][1:])
                    # Use the default system encoding: https://docs.python.org/3/library/functions.html#open
                    with open(args[arg][1:], 'r') as f:  # pylint: disable=unspecified-encoding
                        content = f.read()
                    args[arg] = content
                except IOError:
                    # Leave arg unmodified
                    logger.debug('File Error: Failed to open %s, assume not a file', args[arg][1:])
        return args

    def __init__(self, cli_ctx=None, cli_help=None, **kwargs):
        """ Create the argument parser

        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        :param kwargs: These kwargs are typically used by argparse when creating the subparsers
        """
        from .cli import CLI
        if cli_ctx is not None and not isinstance(cli_ctx, CLI):
            raise CtxTypeError(cli_ctx)
        self.cli_ctx = cli_ctx
        self.cli_help = cli_help
        self.subparsers = {}
        self.parents = kwargs.get('parents', [])
        self.help_file = kwargs.pop('help_file', None)
        # We allow a callable for description to be passed in in order to delay-load any help
        # or description for a command. We better stash it away before handing it off for
        # "normal" argparse handling...
        self._description = kwargs.pop('description', None)
        super().__init__(**kwargs)

    def load_command_table(self, command_loader):
        """ Process the command table and load it into the parser

        :param cmd_tbl: A dictionary containing the commands
        :type cmd_tbl: dict
        """
        cmd_tbl = command_loader.command_table
        grp_tbl = command_loader.command_group_table
        if not cmd_tbl:
            raise ValueError('The command table is empty. At least one command is required.')
        # If we haven't already added a subparser, we
        # better do it.
        if not self.subparsers:
            sp = self.add_subparsers(dest='_command')
            sp.required = True
            self.subparsers = {(): sp}

        for command_name, metadata in cmd_tbl.items():
            subparser = self._get_subparser(command_name.split(), grp_tbl)
            command_verb = command_name.split()[-1]
            # To work around http://bugs.python.org/issue9253, we artificially add any new
            # parsers we add to the "choices" section of the subparser.
            subparser = self._get_subparser(command_name.split(), grp_tbl)
            deprecate_info = metadata.deprecate_info
            if not subparser or (deprecate_info and deprecate_info.expired()):
                continue
            # inject command_module designer's help formatter -- default is HelpFormatter
            fc = metadata.formatter_class or argparse.HelpFormatter

            command_parser = subparser.add_parser(command_verb,
                                                  description=metadata.description,
                                                  parents=self.parents,
                                                  conflict_handler='error',
                                                  help_file=metadata.help,
                                                  formatter_class=fc,
                                                  cli_help=self.cli_help)
            command_parser.cli_ctx = self.cli_ctx
            command_validator = metadata.validator
            argument_validators = []
            argument_groups = {}
            for arg in metadata.arguments.values():

                # don't add deprecated arguments to the parser
                deprecate_info = arg.type.settings.get('deprecate_info', None)
                if deprecate_info and deprecate_info.expired():
                    continue

                if arg.validator:
                    argument_validators.append(arg.validator)
                if arg.arg_group:
                    try:
                        group = argument_groups[arg.arg_group]
                    except KeyError:
                        # group not found so create
                        group_name = '{} Arguments'.format(arg.arg_group)
                        group = command_parser.add_argument_group(arg.arg_group, group_name)
                        argument_groups[arg.arg_group] = group
                    param = CLICommandParser._add_argument(group, arg)
                else:
                    param = CLICommandParser._add_argument(command_parser, arg)
                param.completer = arg.completer
                param.deprecate_info = arg.deprecate_info
                param.preview_info = arg.preview_info
                param.experimental_info = arg.experimental_info
                param.default_value_source = arg.default_value_source
            command_parser.set_defaults(
                func=metadata,
                command=command_name,
                _command_validator=command_validator,
                _argument_validators=argument_validators,
                _parser=command_parser)

    def _get_subparser(self, path, group_table=None):
        """For each part of the path, walk down the tree of
        subparsers, creating new ones if one doesn't already exist.
        """
        group_table = group_table or {}
        for length in range(0, len(path)):
            parent_path = path[:length]
            parent_subparser = self.subparsers.get(tuple(parent_path), None)
            if not parent_subparser:
                # No subparser exists for the given subpath - create and register
                # a new subparser.
                # Since we know that we always have a root subparser (we created)
                # one when we started loading the command table, and we walk the
                # path from left to right (i.e. for "cmd subcmd1 subcmd2", we start
                # with ensuring that a subparser for cmd exists, then for subcmd1,
                # subcmd2 and so on), we know we can always back up one step and
                # add a subparser if one doesn't exist
                command_group = group_table.get(' '.join(parent_path))
                if command_group:
                    deprecate_info = command_group.group_kwargs.get('deprecate_info', None)
                    if deprecate_info and deprecate_info.expired():
                        continue
                grandparent_path = path[:length - 1]
                grandparent_subparser = self.subparsers[tuple(grandparent_path)]
                new_path = path[length - 1]
                new_parser = grandparent_subparser.add_parser(new_path, cli_help=self.cli_help)

                # Due to http://bugs.python.org/issue9253, we have to give the subparser
                # a destination and set it to required in order to get a meaningful error
                parent_subparser = new_parser.add_subparsers(dest='_subcommand')
                command_group = group_table.get(' '.join(parent_path), None)
                deprecate_info = None
                if command_group:
                    deprecate_info = command_group.group_kwargs.get('deprecate_info', None)
                parent_subparser.required = True
                parent_subparser.deprecate_info = deprecate_info
                self.subparsers[tuple(path[0:length])] = parent_subparser
        return parent_subparser

    def validation_error(self, message):
        return super().error(message)

    def is_group(self):
        """ Determine if this parser instance represents a group
            or a command. Anything that has a func default is considered
            a group. This includes any dummy commands served up by the
            "filter out irrelevant commands based on argv" command filter """
        cmd = self._defaults.get('func', None)
        return not (cmd and cmd.handler)

    def __getattribute__(self, name):
        """ Since getting the description can be expensive (require module loads), we defer
            this until someone actually wants to use it (i.e. show help for the command)
        """
        if name == 'description':
            if self._description:
                self.description = self._description() \
                    if callable(self._description) else self._description
                self._description = None
        return object.__getattribute__(self, name)

    def format_help(self):
        is_group = self.is_group()
        self.cli_help.show_help(self.prog.split()[0],
                                self.prog.split()[1:],
                                self._actions[-1] if is_group else self,
                                is_group)
        self.exit()

    def parse_args(self, args=None, namespace=None):
        """ Overrides argparse.ArgumentParser.parse_args

        Enables '@'-prefixed files to be expanded before arguments are processed
        by ArgumentParser.parse_args as usual
        """
        self._expand_prefixed_files(args)
        return super().parse_args(args)

    def _check_value(self, action, value):
        # Override to customize the error message when a argument is not among the available choices
        # converted value must be one of the choices (if specified)
        import difflib
        import sys

        if action.choices is not None and value not in action.choices:
            if action.dest in ["_command", "_subcommand"]:
                # Command
                error_msg = "{prog}: '{value}' is not in the '{prog}' command group. See '{prog} --help'.".format(
                    prog=self.prog, value=value)
                logger.error(error_msg)
                # Show suggestions
                candidates = difflib.get_close_matches(value, action.choices, cutoff=0.7)
                if candidates:
                    suggestion_msg = "\nThe most similar choices to '{value}':\n".format(value=value)
                    suggestion_msg += '\n'.join(['\t' + candidate for candidate in candidates])
                    print(suggestion_msg, file=sys.stderr)
            else:
                # Argument
                error_msg = "{prog}: '{value}' is not a valid value for '{name}'.".format(
                    prog=self.prog, value=value,
                    name=argparse._get_action_name(action))  # pylint: disable=protected-access
                logger.error(error_msg)
                # Show all allowed values
                suggestion_msg = "Allowed values: " + ', '.join(action.choices)
                print(suggestion_msg, file=sys.stderr)
            self.exit(2)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import unittest
from .const import ENV_LIVE_TEST


def live_only():
    return unittest.skipUnless(
        os.environ.get(ENV_LIVE_TEST, False),
        'This is a live only test. A live test will bypass all vcrpy components.')


def record_only():
    return unittest.skipUnless(
        not os.environ.get(ENV_LIVE_TEST, False),
        'This test is excluded from being run live. To force a recording, please remove the recording file.')
